from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class DiagnosticResourceKeyVault(AzureResources):

    def __init__(self, facade: AzureFacade, resource_id: str, subscription_id: str):
        super().__init__(facade)
        self.resource_id = resource_id
        self.subscription_id = subscription_id

    async def fetch_all(self):
        diagnostic_settings = await self.facade.loggingmonitoring.get_diagnostic_settings(self.subscription_id,
                                                                                          self.resource_id)
        self._parse_diagnostic_settings(diagnostic_settings)

    def _parse_diagnostic_settings(self, diagnostic_settings):
        self.update({
            'audit_event_enabled': self.ensure_audit_event_enabled(diagnostic_settings)
        })

    def ensure_audit_event_enabled(self, diagnostic_settings):
        for diagnostic_setting in diagnostic_settings:
            for log in diagnostic_setting.logs:
                if log.category == 'AuditEvent' and log.enabled and log.retention_policy.days > 0:
                    return True
        return False



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id


class BuildProjects(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super().__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_projects = await self.facade.codebuild.get_projects(self.region)
        for list_raw_project in raw_projects:
            for raw_project in list_raw_project.get('projects'):
                id, build_project = self._parse_build_projects(raw_project)
                self[id] = build_project

    def _parse_build_projects(self, raw_build_project):
        project_dict = {}
        project_dict['id'] = raw_build_project.get('arn')
        project_dict['arn'] = raw_build_project.get('arn')
        project_dict['name'] = raw_build_project.get('name')
        if 'vpcConfig' in raw_build_project:
            project_dict['vpc'] = raw_build_project.get('vpcConfig').get('vpcId')
            project_dict['subnets'] = raw_build_project.get('vpcConfig').get('subnets')
            project_dict['security_groups'] = raw_build_project.get('vpcConfig').get('securityGroupIds')
        return project_dict['id'], project_dict



from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions

from .web_apps import WebApplication


class AppServices(Subscriptions):
    _children = [
        (WebApplication, 'web_apps')
    ]



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .alarms import Alarms
from .metric_filters import MetricFilters


class CloudWatch(Regions):
    _children = [
        (Alarms, 'alarms'),
        (MetricFilters, 'metric_filters')
    ]

    def __init__(self, facade: AWSFacade):
        super().__init__('cloudwatch', facade)

    async def finalize(self):

        # For each region, check if at least one metric filter covers the desired events
        for region in self['regions']:
            self['regions'][region]['metric_filters_pattern_checks'] = {}
            # Initialize results at "False"
            self['regions'][region]['metric_filters_pattern_checks']['unauthorized_api_calls'] = False
            self['regions'][region]['metric_filters_pattern_checks']['console_login_mfa'] = False
            self['regions'][region]['metric_filters_pattern_checks']['root_usage'] = False
            self['regions'][region]['metric_filters_pattern_checks']['iam_policy_changes'] = False
            self['regions'][region]['metric_filters_pattern_checks']['cloudtrail_configuration_changes'] = False
            self['regions'][region]['metric_filters_pattern_checks']['console_authentication_failures'] = False
            self['regions'][region]['metric_filters_pattern_checks']['cmk_deletion'] = False
            self['regions'][region]['metric_filters_pattern_checks']['s3_policy_changes'] = False
            self['regions'][region]['metric_filters_pattern_checks']['aws_configuration_changes'] = False
            self['regions'][region]['metric_filters_pattern_checks']['security_group_changes'] = False
            self['regions'][region]['metric_filters_pattern_checks']['nacl_changes'] = False
            self['regions'][region]['metric_filters_pattern_checks']['network_gateways_changes'] = False
            self['regions'][region]['metric_filters_pattern_checks']['route_table_changes'] = False
            self['regions'][region]['metric_filters_pattern_checks']['vpc_changes'] = False
            for metric_filter_id, metric_filter in self['regions'][region]['metric_filters'].items():
                # Check events
                if metric_filter['pattern'] == "{ ($.errorCode = \"*UnauthorizedOperation\") || ($.errorCode = \"AccessDenied*\") }":
                    self['regions'][region]['metric_filters_pattern_checks']['unauthorized_api_calls'] = True
                if metric_filter['pattern'] == "{ ($.eventName = \"ConsoleLogin\") && ($.additionalEventData.MFAUsed != \"Yes\") }":
                    self['regions'][region]['metric_filters_pattern_checks']['console_login_mfa'] = True
                if metric_filter['pattern'] == "{ $.userIdentity.type = \"Root\" && $.userIdentity.invokedBy NOT EXISTS && $.eventType != \"AwsServiceEvent\" }":
                    self['regions'][region]['metric_filters_pattern_checks']['root_usage'] = True
                if metric_filter['pattern'] == "{ ($.eventName=DeleteGroupPolicy) || ($.eventName=DeleteRolePolicy) || ($.eventName=DeleteUserPolicy) || ($.eventName=PutGroupPolicy) || ($.eventName=PutRolePolicy) || ($.eventName=PutUserPolicy) || ($.eventName=CreatePolicy) || ($.eventName=DeletePolicy) || ($.eventName=CreatePolicyVersion) || ($.eventName=DeletePolicyVersion) || ($.eventName=AttachRolePolicy) || ($.eventName=DetachRolePolicy) || ($.eventName=AttachUserPolicy) || ($.eventName=DetachUserPolicy) || ($.eventName=AttachGroupPolicy) || ($.eventName=DetachGroupPolicy) }":
                    self['regions'][region]['metric_filters_pattern_checks']['iam_policy_changes'] = True
                if metric_filter['pattern'] == "{ ($.eventName = CreateTrail) || ($.eventName = UpdateTrail) || ($.eventName = DeleteTrail) || ($.eventName = StartLogging) || ($.eventName = StopLogging) }":
                    self['regions'][region]['metric_filters_pattern_checks']['cloudtrail_configuration_changes'] = True
                if metric_filter['pattern'] == "{ ($.eventName = ConsoleLogin) && ($.errorMessage = \"Failed authentication\") }":
                    self['regions'][region]['metric_filters_pattern_checks']['console_authentication_failures'] = True
                if metric_filter['pattern'] == "{ ($.eventSource = kms.amazonaws.com) && (($.eventName = DisableKey) || ($.eventName = ScheduleKeyDeletion)) }":
                    self['regions'][region]['metric_filters_pattern_checks']['cmk_deletion'] = True
                if metric_filter['pattern'] == "{ ($.eventSource = s3.amazonaws.com) && (($.eventName = PutBucketAcl) || ($.eventName = PutBucketPolicy) || ($.eventName = PutBucketCors) || ($.eventName = PutBucketLifecycle) || ($.eventName = PutBucketReplication) || ($.eventName = DeleteBucketPolicy) || ($.eventName = DeleteBucketReplication)) }":
                    self['regions'][region]['metric_filters_pattern_checks']['s3_policy_changes'] = True
                if metric_filter['pattern'] == "{ ($.eventSource = config.amazonaws.com) && (($.eventName = StopConfigurationRecorder) || ($.eventName = DeleteDeliveryChannel) || ($.eventName = PutDeliveryChannel) || ($.eventName = PutConfigurationRecorder)) }":
                    self['regions'][region]['metric_filters_pattern_checks']['aws_configuration_changes'] = True
                if metric_filter['pattern'] == "{ ($.eventName = AuthorizeSecurityGroupIngress) || ($.eventName = AuthorizeSecurityGroupEgress) || ($.eventName = RevokeSecurityGroupIngress) || ($.eventName = RevokeSecurityGroupEgress) || ($.eventName = CreateSecurityGroup) || ($.eventName = DeleteSecurityGroup) }":
                    self['regions'][region]['metric_filters_pattern_checks']['security_group_changes'] = True
                if metric_filter['pattern'] == "{ ($.eventName = CreateNetworkAcl) || ($.eventName = CreateNetworkAclEntry) || ($.eventName = DeleteNetworkAcl) || ($.eventName = DeleteNetworkAclEntry) || ($.eventName = ReplaceNetworkAclEntry) || ($.eventName = ReplaceNetworkAclAssociation) }":
                    self['regions'][region]['metric_filters_pattern_checks']['nacl_changes'] = True
                if metric_filter['pattern'] == "{ ($.eventName = CreateCustomerGateway) || ($.eventName = DeleteCustomerGateway) || ($.eventName = AttachInternetGateway) || ($.eventName = CreateInternetGateway) || ($.eventName = DeleteInternetGateway) || ($.eventName = DetachInternetGateway) }":
                    self['regions'][region]['metric_filters_pattern_checks']['network_gateways_changes'] = True
                if metric_filter['pattern'] == "{ ($.eventName = CreateRoute) || ($.eventName = CreateRouteTable) || ($.eventName = ReplaceRoute) || ($.eventName = ReplaceRouteTableAssociation) || ($.eventName = DeleteRouteTable) || ($.eventName = DeleteRoute) || ($.eventName = DisassociateRouteTable) }":
                    self['regions'][region]['metric_filters_pattern_checks']['route_table_changes'] = True
                if metric_filter['pattern'] == "{ ($.eventName = CreateVpc) || ($.eventName = DeleteVpc) || ($.eventName = ModifyVpcAttribute) || ($.eventName = AcceptVpcPeeringConnection) || ($.eventName = CreateVpcPeeringConnection) || ($.eventName = DeleteVpcPeeringConnection) || ($.eventName = RejectVpcPeeringConnection) || ($.eventName = AttachClassicLinkVpc) || ($.eventName = DetachClassicLinkVpc) || ($.eventName = DisableVpcClassicLink) || ($.eventName = EnableVpcClassicLink) }":
                    self['regions'][region]['metric_filters_pattern_checks']['vpc_changes'] = True
                   


from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class RoleAssignments(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_role_assignment in await self.facade.rbac.get_role_assignments(self.subscription_id):
            id, role_assignment = self._parse_role_assignment(raw_role_assignment)
            self[id] = role_assignment

    def _parse_role_assignment(self, raw_role_assignment):
        role_assignment_dict = {}
        role_assignment_dict['id'] = raw_role_assignment.name
        role_assignment_dict['name'] = raw_role_assignment.name
        role_assignment_dict['role_definition_id'] = raw_role_assignment.role_definition_id
        role_assignment_dict['type'] = raw_role_assignment.type
        role_assignment_dict['scope'] = raw_role_assignment.scope
        role_assignment_dict['principal_id'] = raw_role_assignment.principal_id
        role_assignment_dict['principal_type'] = raw_role_assignment.principal_type
        role_assignment_dict['can_delegate'] = "None"
        role_assignment_dict['additional_properties'] = raw_role_assignment.additional_properties
        return role_assignment_dict['id'], role_assignment_dict



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class BlobServices(AzureResources):

    def __init__(self, facade: AzureFacade, resource_group_name: str, storage_account_name: str, subscription_id: str):
        super().__init__(facade)
        self.resource_group_name = resource_group_name
        self.storage_account_name = storage_account_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        raw_blob_services = await self.facade.storageaccounts.get_blob_services(self.resource_group_name,
                                                                                    self.storage_account_name,
                                                                                    self.subscription_id)
        for raw_blob_service in raw_blob_services:
            id, blob_service = self._parse_blob_service(raw_blob_service)
            self[id] = blob_service

    def _parse_blob_service(self, raw_blob_service):
        blob_service = {}
        blob_service['id'] = get_non_provider_id(raw_blob_service.id.lower())
        blob_service['name'] = raw_blob_service.name
        blob_service['soft_delete_enabled'] = raw_blob_service.delete_retention_policy.enabled

        return blob_service['id'], blob_service



from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.core.console import print_exception
from ScoutSuite.utils import get_user_agent
from azure.mgmt.resource import ResourceManagementClient


class ResourceManagementFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = ResourceManagementClient(self.credentials.get_credentials(),
                                          subscription_id=subscription_id,
                                          user_agent=get_user_agent())
        return client

    async def get_specific_type_resources_with_filter(self, subscription_id: str, resource_type_filter: str):
        try:
            type_filter = " and ".join([
                f'resourceType eq \'{resource_type_filter}\''
            ])
            client = self.get_client(subscription_id)
            resource = await run_concurrently(
                lambda: list(client.resources.list(filter=type_filter))
            )
            return resource
        except Exception as e:
            print_exception(f'Failed to retrieve key vault resources: {e}')
            return []

    async def get_all_resources(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            resource = await run_concurrently(
                lambda: list(client.resources.list())
            )
            return resource
        except Exception as e:
            print_exception(f'Failed to retrieve resources: {e}')
            return []



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.providers.azure.utils import get_resource_group_name

from ScoutSuite.providers.azure.utils import get_resource_group_name

class Instances(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_instance in await self.facade.virtualmachines.get_instances(self.subscription_id):
            id, instance = await self._parse_instance(raw_instance)
            self[id] = instance

    async def _parse_instance(self, raw_instance):
        instance_dict = {}

        instance_dict['id'] = get_non_provider_id(raw_instance.id.lower())
        instance_dict['name'] = raw_instance.name
        instance_dict['vm_id'] = raw_instance.vm_id
        instance_dict['zones'] = raw_instance.zones
        if raw_instance.availability_set is not None:
            #Get the resource group and availability set if set
            try:
                instance_dict['availability_set'] = raw_instance.availability_set.id.split('/')[4] + ':' + raw_instance.availability_set.id.split('/')[8]
            except Exception as e:
                instance_dict['availability_set'] = raw_instance.availability_set.id
        else:
            instance_dict['availability_set'] = None
        if raw_instance.proximity_placement_group is not None:
            #Get the resource group and proximity placement group if set  
            try:
                instance_dict['proximity_placement_group'] = raw_instance.proximity_placement_group.id.split('/')[4] + ':' + raw_instance.proximity_placement_group.id.split('/')[8]
            except Exception as e:
                instance_dict['proximity_placement_group'] = raw_instance.proximity_placement_group.id
        else:
            instance_dict['proximity_placement_group'] = None
        instance_dict['additional_properties'] = list(raw_instance.additional_properties)
        instance_dict['location'] = raw_instance.location
        instance_dict['type'] = raw_instance.type
        instance_dict['resources'] = raw_instance.resources
        if raw_instance.tags is not None:
            instance_dict['tags'] = ["{}:{}".format(key, value) for key, value in  raw_instance.tags.items()]
        else:
            instance_dict['tags'] = []
        instance_dict['resource_group_name'] = get_resource_group_name(raw_instance.id)
        instance_dict['provisioning_state'] = raw_instance.provisioning_state
        if raw_instance.plan is not None:
            instance_dict['plan'] = raw_instance.plan.name
        else:
            instance_dict['plan'] = None
        instance_dict['identity'] = raw_instance.identity

        if raw_instance.additional_capabilities is not None:
            #Get all the enabled additional capabilities ignoring not set or empty values
            instance_dict['additional_capabilities'] = [additional_capability for additional_capability, value in raw_instance.additional_capabilities.__dict__.items() if (value and value is not None)]
        else:
            instance_dict['additional_capabilities'] = None
        instance_dict['license_type'] = raw_instance.license_type

        # TODO process and display the below
        instance_dict['hardware_profile'] = raw_instance.hardware_profile.vm_size
        
        # Handle VMs without diagnostics profile configured
        if raw_instance.diagnostics_profile is not None:
            instance_dict['diagnostics_profile'] = {'Boot Diagnostics': True if raw_instance.diagnostics_profile.boot_diagnostics.enabled else None}
        
        instance_dict['os_profile'] = {}
        if raw_instance.os_profile is not None:
            instance_dict['os_profile']['Administrator Username'] = raw_instance.os_profile.admin_username
            instance_dict['os_profile']['Allow Extension Operations'] = raw_instance.os_profile.allow_extension_operations
            instance_dict['os_profile']['Computer Name'] = raw_instance.os_profile.computer_name
            instance_dict['os_profile']['Custom Data'] = raw_instance.os_profile.custom_data
            instance_dict['os_profile']['Secrets'] = ''.join(raw_instance.os_profile.secrets)
            if raw_instance.os_profile.windows_configuration:
                instance_dict['os_profile']['Unnatended Content'] = raw_instance.os_profile.windows_configuration.additional_unattend_content
                instance_dict['os_profile']['Automatic Updates'] = raw_instance.os_profile.windows_configuration.enable_automatic_updates
                instance_dict['os_profile']['VM Agent Provision'] = raw_instance.os_profile.windows_configuration.provision_vm_agent
                instance_dict['os_profile']['Windows Remote Management'] = raw_instance.os_profile.windows_configuration.win_rm
            elif raw_instance.os_profile.linux_configuration:
                instance_dict['os_profile']['Disable Password Authentication'] = raw_instance.os_profile.linux_configuration.disable_password_authentication
                instance_dict['os_profile']['VM Agent Provision'] = raw_instance.os_profile.linux_configuration.provision_vm_agent

        if raw_instance.storage_profile is not None:
            instance_dict['storage_profile'] = {}
            if raw_instance.storage_profile.image_reference is not None:
                instance_dict['storage_profile']['Publisher'] = raw_instance.storage_profile.image_reference.publisher
            instance_dict['storage_profile']['Release'] = raw_instance.storage_profile.image_reference.version
            instance_dict['storage_profile']['SKU'] = raw_instance.storage_profile.image_reference.sku
            instance_dict['storage_profile']['Offer'] = raw_instance.storage_profile.image_reference.offer
            instance_dict['storage_profile']['Exact Version'] = raw_instance.storage_profile.image_reference.exact_version
            instance_dict['storage_profile']['OS Disk Size (GB)'] = raw_instance.storage_profile.os_disk.disk_size_gb
            instance_dict['storage_profile']['OS Disk Name'] = raw_instance.storage_profile.os_disk.name
            instance_dict['storage_profile']['OS Disk VHD'] = raw_instance.storage_profile.os_disk.vhd
            if raw_instance.storage_profile.os_disk.managed_disk:
                instance_dict['storage_profile'][
                    'OS Managed Disk ID'] = raw_instance.storage_profile.os_disk.managed_disk.id.split('/')[-1]
                instance_dict['storage_profile'][
                    'OS Managed Disk Storage Account Type'] = raw_instance.storage_profile.os_disk.managed_disk.storage_account_type
            else:
                instance_dict['storage_profile']['OS Managed Disk ID'] = None
                instance_dict['storage_profile']['OS Managed Disk Storage Account Type'] = None
            if raw_instance.storage_profile.data_disks is not None and raw_instance.storage_profile.data_disks:
                instance_dict['storage_profile']['Data Disks'] = ["{} ({}GB)".format(disk.name, disk.disk_size_gb) for disk in raw_instance.storage_profile.data_disks]
        else:
            instance_dict['storage_profile'] = None

        instance_dict['network_interfaces'] = []
        for interface in raw_instance.network_profile.network_interfaces:
            instance_dict['network_interfaces'].append(get_non_provider_id(interface.id))

        instance_dict['extensions'] = await self.facade.virtualmachines.get_instance_extensions(
            subscription_id=self.subscription_id,
            instance_name=instance_dict['name'],
            resource_group=get_resource_group_name(raw_instance.id))

        instance_dict['extension_names'] = [extension.name for extension in instance_dict['extensions']]

        return instance_dict['id'], instance_dict



from ScoutSuite.providers.gcp.resources.gke.clusters import Clusters
from ScoutSuite.providers.gcp.resources.projects import Projects


class GKE(Projects):
    _children = [
        (Clusters, 'clusters')
    ]


import subprocess
import unittest
from unittest import mock

import pytest
from ScoutSuite.__main__ import run_from_cli
from ScoutSuite.core.console import set_logger_configuration


class TestScoutSuiteClass(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger_configuration(is_debug=True)
        cls.has_run_scout_suite = False

    @pytest.mark.xfail("only runs with AWS, cannot be used dynamically")
    @staticmethod
    def call_scout_suite(args):
        args = ['./scout.py'] + args

        args.append('aws')

        if TestScoutSuiteClass.profile_name:
            args.append('--profile')
            args.append(TestScoutSuiteClass.profile_name)
        # TODO: FIXME this only tests AWS

        args.append('--force')
        args.append('--debug')
        args.append('--no-browser')
        if TestScoutSuiteClass.has_run_scout_suite:
            args.append('--local')
        TestScoutSuiteClass.has_run_scout_suite = True

        sys = None
        with mock.patch.object(sys, 'argv', args):
            return run_from_cli()

    def test_scout_suite_help(self):
        """Make sure that ScoutSuite does not crash with --help"""
        command = './scout.py --help'
        process = subprocess.Popen(command, shell=True, stdout=None)
        process.wait()
        assert process.returncode == 0

    @pytest.mark.xfail
    def test_scout_suite_default_run(self):
        """Make sure that ScoutSuite's default run does not crash"""
        rc = self.call_scout_suite([])
        assert (rc == 0)



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.utils import get_non_provider_id


class UptimeChecks(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_uptime_checks = await self.facade.stackdrivermonitoring.get_uptime_checks(self.project_id)
        for raw_uptime_check in raw_uptime_checks:
            uptime_check_name, uptime_check = self._parse_uptime_check(raw_uptime_check)
            self[uptime_check_name] = uptime_check

    def _parse_uptime_check(self, raw_uptime_check):
        uptime_check_dict = {}
        uptime_check_dict['id'] = get_non_provider_id(raw_uptime_check.name)
        uptime_check_dict['name'] = raw_uptime_check.display_name
        uptime_check_dict['monitored_resource'] = raw_uptime_check.monitored_resource
        uptime_check_dict['http_check'] = raw_uptime_check.http_check
        uptime_check_dict['period'] = raw_uptime_check.period
        uptime_check_dict['timeout'] = raw_uptime_check.timeout
        return uptime_check_dict['id'], uptime_check_dict



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureCompositeResources
from ScoutSuite.providers.utils import get_non_provider_id
from .diagnostic_resource_key_vault import DiagnosticResourceKeyVault


class Resources(AzureCompositeResources):
    """
    TODO this is commented out since DiagnositcResourceKeyVault.get_diagnostic_settings needs to be fixed
    _children = [
        (DiagnosticResourceKeyVault, 'diagnostic_key_vault'),
    ]"""
    
    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id
    async def fetch_all(self):
        for raw_resource in await self.facade.resourcemanagement.get_specific_type_resources_with_filter(
                self.subscription_id, 'Microsoft.KeyVault/vaults'):
            id, resource = self._parse_resource(raw_resource)
            self[id] = resource
         
        """
        TODO this is commented out since DiagnositcResourceKeyVault.get_diagnostic_settings needs to be fixed
        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={resource_id: {'resource_id': resource['id'],
                                  'subscription_id': self.subscription_id}
                    for (resource_id, resource) in self.items()}
        )
        """

    def _parse_resource(self, raw_resource):
        resource = {}
        resource['id'] = get_non_provider_id(raw_resource.id.lower())
        resource['name'] = raw_resource.name
        if raw_resource.tags is not None:
            resource['tags'] = ["{}:{}".format(key, value) for key, value in raw_resource.tags.items()]
        else:
            resource['tags'] = []
        return resource['id'], resource



from ScoutSuite.providers.azure.resources.base import AzureCompositeResources


class Subscriptions(AzureCompositeResources):

    """This class represents a collection of Azure Resources that are grouped by subscription.
    Classes extending Subscriptions should implement the method _fetch_children() with a subscription ID as paramater.
    The children resources will be stored with the following structure {<subscriptions>: {<subscription_id>: {<child_name>: {<child_id>: <child_instance>}}}}.
    """

    async def fetch_all(self):
        """This method fetches all the Azure subscriptions that can be accessed with the given run configuration.
        It then fetches all the children defined in _children and groups them by subscription.
        """

        raw_subscriptions = await self.facade.get_subscriptions()

        if raw_subscriptions:
            self['subscriptions'] = {subscription.subscription_id: {}
                                     for subscription in raw_subscriptions}
        else:
            self['subscriptions'] = {}
        await self._fetch_children_of_all_resources(
            resources=self['subscriptions'],
            scopes={subscription_id: {'subscription_id': subscription_id} for subscription_id in self['subscriptions']})
        self._set_counts()

    def _set_counts(self):
        for _, child_name in self._children:
            self[child_name + '_count'] = sum([subscription[child_name + '_count']
                                               for subscription in self['subscriptions'].values()])



from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions

from .mysql_servers import MySQLServers


class MySQLServers(Subscriptions):
    _children = [
        (MySQLServers, 'servers')
    ]


#!/usr/bin/env python3

from ScoutSuite.providers.aws.utils import get_caller_identity
from ScoutSuite.core.console import set_logger_configuration, print_info, print_exception
from tools.utils import results_file_to_dict

import datetime
import argparse
import boto3


def upload_findigs_to_securityhub(session, formatted_findings_list):
    try:
        if formatted_findings_list:
            print_info('Batch uploading {} findings'.format(len(formatted_findings_list)))
            securityhub = session.client('securityhub')
            response = securityhub.batch_import_findings(Findings=formatted_findings_list)
            print_info('Upload completed, {} succeeded, {} failed'.format(response.get('SuccessCount'),
                                                                          response.get('FailedCount')))
            return response
    except Exception as e:
        print_exception(f'Unable to upload findings to Security Hub: {e}')


def format_finding_to_securityhub_format(aws_account_id,
                                         region,
                                         creation_date,
                                         finding_key,
                                         finding_value):
    try:

        if finding_value.get('level') == 'danger':
            label = 'HIGH'
        elif finding_value.get('level') == 'warning':
            label = 'MEDIUM'
        else:
            label = 'INFORMATIONAL'

        format_time = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()

        formatted_finding = {
            'SchemaVersion': '2018-10-08',
            'Id': finding_key,
            'ProductArn':
                'arn:aws:securityhub:' + region + ':' + aws_account_id + ':product/' + aws_account_id + '/default',
            'GeneratorId': f'scoutsuite-{aws_account_id}',
            'AwsAccountId': aws_account_id,
            'Types': ['Software and Configuration Checks/AWS Security Best Practices'],
            'FirstObservedAt': creation_date,
            'CreatedAt': format_time,
            'UpdatedAt': format_time,
            'Severity': {
                'Label': label
            },
            'Title': finding_value.get('description'),
            'Description': finding_value.get('rationale') if finding_value.get('rationale') else 'None',
            'Remediation': {
                'Recommendation': {
                    'Text': finding_value.get('remediation', 'None') if finding_value.get('remediation') else 'None'
                }
            },
            'ProductFields': {'Product Name': 'Scout Suite'},
            'Resources': [  # TODO this lacks affected resources
                {
                    'Type': 'AwsAccount',
                    'Id': 'AWS::::Account:' + creation_date,
                    'Partition': 'aws',
                    'Region': region
                }
            ],
            'Compliance': {
                'Status': 'FAILED'
            },
            'RecordState': 'ACTIVE'
        }
        return formatted_finding
    except Exception as e:
        print_exception(f'Unable to process finding: {e}')


def process_results_file(f,
                         region):
    try:
        formatted_findings_list = []
        results = results_file_to_dict(f)

        aws_account_id = results["account_id"]
        creation_date = datetime.datetime.strptime(results["last_run"]["time"], '%Y-%m-%d %H:%M:%S%z').isoformat()

        for service in results.get('service_list'):
            for finding_key, finding_value in results.get('services', {}).get(service).get('findings').items():
                if finding_value.get('items'):
                    formatted_finding = format_finding_to_securityhub_format(aws_account_id,
                                                                             region,
                                                                             creation_date,
                                                                             finding_key,
                                                                             finding_value)
                    formatted_findings_list.append(formatted_finding)

        return formatted_findings_list
    except Exception as e:
        print_exception(f'Unable to process results file: {e}')


def run(profile, file):
    session = boto3.Session(profile_name=profile)
    # Test querying for current user
    get_caller_identity(session)
    print_info(f'Authenticated with profile {profile}')

    try:
        with open(file) as f:
            formatted_findings_list = process_results_file(f,
                                                           session.region_name)
    except Exception as e:
        print_exception(f'Unable to open file {file}: {e}')

    upload_findigs_to_securityhub(session, formatted_findings_list)


if __name__ == "__main__":

    # Configure the debug level
    set_logger_configuration()

    parser = argparse.ArgumentParser(description='Tool to upload a JSON report to AWS Security Hub')
    parser.add_argument('-p', '--profile',
                        required=False,
                        default="default",
                        help="The named profile to use to authenticate to AWS. Defaults to \"default\".")
    parser.add_argument('-f', '--file',
                        required=True,
                        help="The path of the JSON results file to process, e.g. "
                             "\"scoutsuite-report/scoutsuite-results/scoutsuite_results_aws-<profile>.js\".")
    args = parser.parse_args()

    try:
        run(args.profile, args.file)
    except Exception as e:
        print_exception(f'Unable to complete: {e}')



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class LogProfiles(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for log_profile in await self.facade.loggingmonitoring.get_log_profiles(self.subscription_id):
            id, log_profiles = self._parse_log_profile(log_profile)
            self[id] = log_profiles

    def _parse_log_profile(self, log_profile):
        log_profile_dict = {}

        log_profile_dict['id'] = get_non_provider_id(log_profile.id.lower())
        log_profile_dict['name'] = log_profile.name
        log_profile_dict['storage_account_id'] = log_profile.storage_account_id
        log_profile_dict['service_bus_rule_id'] = log_profile.service_bus_rule_id
        log_profile_dict['retention_policy_enabled'] = log_profile.retention_policy.enabled
        log_profile_dict['retention_policy_days'] = log_profile.retention_policy.days
        log_profile_dict['captures_all_activities'] = self.profile_captures_all_activities(log_profile)

        return log_profile_dict['id'], log_profile_dict

    def profile_captures_all_activities(self, log_profile):
        categories = log_profile.categories
        if 'Delete' in categories and 'Write' in categories and 'Action' in categories:
            return True
        return False




from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class BindingsSeparationDuties(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_bindings = await self.facade.cloudresourcemanager.get_member_bindings(self.project_id)
        binding_id, binding = await self._parse_binding_separation(raw_bindings)
        self[binding_id] = binding

    async def _parse_binding_separation(self, raw_bindings):
        binding_dict = {}
        binding_dict['id'] = self.project_id
        binding_dict['name'] = self.project_id
        binding_dict["account_separation_duties"] = self.ensure_seperation_duties(raw_bindings)
        binding_dict["kms_separation_duties"] = self.ensure_KMS_seperation_duties(raw_bindings)

        return binding_dict['id'], binding_dict

    def ensure_seperation_duties(self, raw_bindings):
        # This function checks if a member has both the iam.serviceAccountAdmin role and iam.serviceAccountUser role.
        # If the roles do have a common member the function returns False
        list_members_role_admin = []
        list_members_role_other = []
        for binding in raw_bindings:
            role = binding['role'].split('/')[-1]
            if role == 'iam.serviceAccountAdmin':
                list_members_role_admin = binding['members']
            if role == 'iam.serviceAccountUser':
                list_members_role_other = binding['members']

        common_members = list(set(list_members_role_admin).intersection(list_members_role_other))
        if common_members:
            return False
        return True

    def ensure_KMS_seperation_duties(self, raw_bindings):
        # This function checks if a member has both the cloudkms.admin role and either
        # cloudkms.cryptoKeyEncrypterDecrypter, cloudkms.cryptoKeyEncrypter, cloudkms.cryptoKeyDecrypter role.
        # If the roles do have a common member the function returns False
        list_members_role_admin = []
        list_members_role_others = {"cloudkms.cryptoKeyEncrypterDecrypter": [],
                                    "cloudkms.cryptoKeyEncrypter": [],
                                    "cloudkms.cryptoKeyDecrypter": []}
        for binding in raw_bindings:
            role = binding['role'].split('/')[-1]
            if role == 'cloudkms.admin':
                list_members_role_admin = binding['members']
            if role == 'cloudkms.cryptoKeyEncrypterDecrypter':
                list_members_role_others['cloudkms.cryptoKeyEncrypterDecrypter'] = binding['members']
            if role == 'cloudkms.cryptoKeyEncrypter':
                list_members_role_others['cloudkms.cryptoKeyEncrypter'] = binding['members']
            if role == 'cloudkms.cryptoKeyDecrypter':
                list_members_role_others['cloudkms.cryptoKeyDecrypter'] = binding['members']

        common_members1 = list(
            set(list_members_role_admin).intersection(list_members_role_others['cloudkms.cryptoKeyEncrypterDecrypter']))
        common_members2 = list(
            set(list_members_role_admin).intersection(list_members_role_others['cloudkms.cryptoKeyEncrypter']))
        common_members3 = list(
            set(list_members_role_admin).intersection(list_members_role_others['cloudkms.cryptoKeyDecrypter']))
        if common_members1 or common_members2 or common_members3:
            return False
        return True



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.basefacade import GCPBaseFacade
from ScoutSuite.providers.gcp.facade.utils import GCPFacadeUtils
from ScoutSuite.providers.utils import map_concurrently, run_concurrently


class BigQueryFacade(GCPBaseFacade):
    def __init__(self):

        super().__init__('bigquery', 'v2')  # API Client

    async def get_datasets(self, project_id: str):
        try:
            bigquery_client = self._get_client()
            datasets = bigquery_client.datasets()

            # get list of datasets
            request = datasets.list(projectId=project_id)
            results = await GCPFacadeUtils.get_all('datasets', request, datasets)
            # extract ids
            dataset_ids = [dataset.get('id').split(':')[-1] for dataset in results]
        except Exception as e:
            print_exception(f'Failed to list BigQuery datasets: {e}')
            return []
        else:
            return await map_concurrently(self._get_dataset, dataset_ids, project_id=project_id)

    async def _get_dataset(self, dataset_id: str, project_id: str):
        try:
            bigquery_client = self._get_client()
            datasets = bigquery_client.datasets()
            request = datasets.get(projectId=project_id, datasetId=dataset_id)
            return await run_concurrently(
                lambda: request.execute()
            )
        except Exception as e:
            print_exception(f'Failed to retrieve BigQuery datasets {dataset_id}: {e}')
            return {}



import logging
import enum

from google.auth.credentials import Credentials as GCPCredentials
from kubernetes import config, client
from ScoutSuite.providers.base.authentication_strategy import AuthenticationStrategy
from ScoutSuite.providers.aws.authentication_strategy import AWSAuthenticationStrategy, AWSCredentials
from ScoutSuite.providers.azure.authentication_strategy import AzureAuthenticationStrategy, AzureCredentials
from ScoutSuite.providers.gcp.authentication_strategy import GCPAuthenticationStrategy


class KubernetesCredentials:
    def __init__(self) -> None:
        self.cluster_provider: str = None
        self.cluster_context: str = None
        self.api_client: client.ApiClient = None
        self.fetch_local: bool = False
        
        self.aws: AWSCredentials = None
        self.azure: AzureCredentials = None
        self.gcp: GCPCredentials = None

class ClusterProvider(enum.Enum):
    # Azure
    AKS = 'aks'
    # AWS
    EKS = 'eks'
    # GCP
    GKE = 'gke'

class ResourceTemplates(enum.Enum):
    RESOURCE_CONTAINERS = 'kubernetes_resource_containers'
    RESOURCE_HOST = 'kubernetes_resource_host'

class KubernetesAuthenticationStrategy(AuthenticationStrategy):
    '''
    Implements authentication for the Kubernetes provider.
    '''

    def authenticate(self, **kwargs):
        '''Obtain credentials to interact with the Kubernetes cluster'''

        logging.getLogger('kubernetes.client.rest').setLevel(logging.ERROR)

        cluster_provider = kwargs.get('kubernetes_cluster_provider')
        config_file = kwargs.get('kubernetes_config_file')
        context = kwargs.get('kubernetes_context')
        persist_config = kwargs.get('kubernetes_persist_config')
        fetch_local = kwargs.get('kubernetes_fetch_local')

        credentials = KubernetesCredentials()
        if cluster_provider in [ClusterProvider.AKS.value]:
            subscription_id = subscription_id=kwargs.get('kubernetes_azure_subscription_id')
            credentials.azure = AzureAuthenticationStrategy().authenticate(cli=True, subscription_id=subscription_id)

        elif cluster_provider in [ClusterProvider.EKS.value]:
            credentials.aws = AWSAuthenticationStrategy().authenticate()

        elif cluster_provider in [ClusterProvider.GKE.value]:
            credentials.gcp = GCPAuthenticationStrategy().authenticate(user_account=True)

        config.load_kube_config(config_file, context, None, persist_config)

        credentials.cluster_provider = cluster_provider
        credentials.cluster_context = context or config.list_kube_config_contexts(config_file)[1]['context']['cluster']
        credentials.api_client = config.new_client_from_config(config_file, context, persist_config)
        credentials.fetch_local = fetch_local

        return credentials



import re

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.base import GCPBaseFacade
from ScoutSuite.providers.utils import run_concurrently, get_and_set_concurrently


class GKEFacade(GCPBaseFacade):
    def __init__(self, gce_facade):
        super(GKEFacade, self).__init__('container', 'v1beta1')
        self._gce_facade = gce_facade

    async def get_clusters(self, project_id):
        try:
            gke_client = self._get_client()
            response = await run_concurrently(
                lambda: gke_client.projects().locations().clusters().list(parent=f"projects/{project_id}/locations/-").execute()
            )
            clusters = response.get('clusters', [])
            await get_and_set_concurrently([self._get_and_set_private_google_access_enabled],
                                           clusters, project_id=project_id)
            return clusters
        except Exception as e:
            print_exception('Failed to retrieve clusters: {}'.format(e))
            return []

    async def _get_and_set_private_google_access_enabled(self, cluster, project_id):
        try:
            region = self._get_cluster_region(cluster)
            subnetwork_project_id = self._get_cluster_subnetwork_project(cluster)
            subnetwork = await self._gce_facade.get_subnetwork(subnetwork_project_id, region, cluster['subnetwork'])
            if subnetwork:
                cluster['privateIpGoogleAccess'] = subnetwork.get('privateIpGoogleAccess')
            else:
                cluster['privateIpGoogleAccess'] = None
        except Exception as e:
            print_exception('Failed to retrieve cluster private IP Google access config: {}'.format(e))
            cluster['privateIpGoogleAccess'] = None

    # The cluster location is given as <region>-<zone>. See the the following link for more info: 
    # https://cloud.google.com/compute/docs/regions-zones/#identifying_a_region_or_zone
    def _get_cluster_region(self, cluster):
        region_regex = re.compile("^([\\w]+-[\\w]+)")
        result = region_regex.search(cluster['location'])
        return result.group(1)

    # Subnetwork can be in different project
    # networkConfig.subnetwork is like projects/{project}/regions/{region}/subnetworks/{subnetworkname}
    def _get_cluster_subnetwork_project(self, cluster):
        return cluster['networkConfig']['subnetwork'].split('/')[1]



from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions
from ScoutSuite.providers.azure.resources.loggingmonitoring.log_profiles import LogProfiles
from ScoutSuite.providers.azure.resources.loggingmonitoring.diagnostic_settings import DiagnosticSettings
from ScoutSuite.providers.azure.resources.loggingmonitoring.activity_log_alerts import ActivityLogAlerts
from ScoutSuite.providers.azure.resources.loggingmonitoring.resources import Resources


class AKS(Subscriptions):
    _children = [
        (LogProfiles, 'log_profiles'),
        (DiagnosticSettings, 'diagnostic_settings'),
        (ActivityLogAlerts, 'log_alerts'),
        (Resources, 'resources_logging')
    ]



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class ActivityLogAlerts(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        log_alerts = await self.facade.loggingmonitoring.get_activity_log_alerts(self.subscription_id)
        self[self.subscription_id] = self._parse_log_alerts(log_alerts)

    def _parse_log_alerts(self, log_alerts):
        log_alerts_dict = {}
        log_alerts_dict['create_policy_assignment_exist'] = self.ensure_alert_exist(log_alerts,
                                                                                    'Microsoft.Authorization'
                                                                                    '/policyAssignments/write')
        log_alerts_dict['create_update_NSG_exist'] = self.ensure_alert_exist(log_alerts,
                                                                             'Microsoft.Network/networkSecurityGroups'
                                                                             '/write')
        log_alerts_dict['delete_NSG_exist'] = self.ensure_alert_exist(log_alerts,
                                                                      'Microsoft.Network/networkSecurityGroups/delete')
        log_alerts_dict['create_update_NSG_rule_exist'] = self.ensure_alert_exist(log_alerts,
                                                                                  'Microsoft.Network'
                                                                                  '/networkSecurityGroups'
                                                                                  '/securityRules/write')
        log_alerts_dict['delete_NSG_rule_exist'] = self.ensure_alert_exist(log_alerts,
                                                                           'Microsoft.Network/networkSecurityGroups'
                                                                           '/securityRules/delete')
        log_alerts_dict['create_update_security_solution_exist'] = self.ensure_alert_exist(log_alerts,
                                                                                           'Microsoft.Security'
                                                                                           '/securitySolutions/write')
        log_alerts_dict['delete_security_solution_exist'] = self.ensure_alert_exist(log_alerts,
                                                                                    'Microsoft.Security'
                                                                                    '/securitySolutions/delete')
        log_alerts_dict['create_delete_firewall_rule_exist'] = self.ensure_alert_exist(log_alerts,
                                                                                       'Microsoft.Sql/servers'
                                                                                       '/firewallRules/write')

        return log_alerts_dict

    def ensure_alert_exist(self, log_alerts, equals_value: str):
        for log_alert in log_alerts:
            if log_alert.location == 'Global' and log_alert.enabled:
                if '/subscriptions/' + self.subscription_id in log_alert.scopes:
                    for condition in log_alert.condition.all_of:
                        if condition.field == 'operationName' and condition.equals == equals_value:
                            return True
        return False



from google.cloud import monitoring as stackdrivermonitoring
from google.api_core.gapic_v1.client_info import ClientInfo

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.utils import get_user_agent


class StackdriverMonitoringFacade:
    # TODO find a way to skip the project if it's not configured as a stackdriver workspace

    def get_uptime_client(self):
        client_info = ClientInfo(user_agent=get_user_agent())
        client = stackdrivermonitoring.UptimeCheckServiceClient(client_info=client_info)
        return client

    def get_alerts_client(self):
        client_info = ClientInfo(user_agent=get_user_agent())
        client = stackdrivermonitoring.AlertPolicyServiceClient(client_info=client_info)
        return client

    async def get_uptime_checks(self, project_id: str):
        try:
            client = self.get_uptime_client()
            name = client.project_path(project_id)
            return await run_concurrently(lambda: [r for r in client.list_uptime_check_configs(name)])
        except Exception as e:
            if 'is not a workspace' not in getattr(e, 'message', '') and '404' not in str(e):
                print_exception(f'Failed to retrieve uptime checks: {e}')
            return []

    async def get_alert_policies(self, project_id: str):
        try:
            client = self.get_alerts_client()
            name = client.project_path(project_id)
            return await run_concurrently(lambda: [r for r in client.list_alert_policies(name)])
        except Exception as e:
            if 'is not a workspace' not in getattr(e, 'message', '') and '404' not in str(e):
                print_exception(f'Failed to retrieve alert policies: {e}')
            return []



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class ConfigurationLogRetentionDays(AzureResources):

    def __init__(self, facade: AzureFacade, resource_group_name: str, server_name: str, subscription_id: str):
        super().__init__(facade)
        self.resource_group_name = resource_group_name
        self.server_name = server_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        configuration = await self.facade.postgresqldatabase.get_config(self.resource_group_name, self.server_name,
                                                                        self.subscription_id, 'log_retention_days')
        self._parse_configuration(configuration)

    def _parse_configuration(self, configuration):
        self.update({
            'value': configuration.value
        })



import copy
import os

from ScoutSuite.core.console import print_error, print_exception, print_warning, print_debug
from ScoutSuite.providers.aws.services import AWSServicesConfig
from ScoutSuite.providers.aws.resources.vpc.base import put_cidr_name
from ScoutSuite.providers.aws.utils import ec2_classic, get_aws_account_id, get_partition_name
from ScoutSuite.providers.base.configs.browser import combine_paths, get_object_at, get_value_at
from ScoutSuite.providers.base.provider import BaseProvider
from ScoutSuite.utils import manage_dictionary


class AWSProvider(BaseProvider):
    """
    Implements provider for AWS
    """

    def __init__(self, profile='default', report_dir=None, timestamp=None, services=None, skipped_services=None,
                 result_format='json', **kwargs):
        services = [] if services is None else services
        skipped_services = [] if skipped_services is None else skipped_services

        self.metadata_path = '%s/metadata.json' % os.path.split(os.path.abspath(__file__))[0]

        self.sg_map = {}
        self.subnet_map = {}

        self.profile = profile
        self.services_config = AWSServicesConfig

        self.provider_code = 'aws'
        self.provider_name = 'Amazon Web Services'
        self.environment = self.profile
        self.result_format = result_format

        self.credentials = kwargs['credentials']

        self.partition = get_partition_name(self.credentials.session)

        self.account_id = get_aws_account_id(self.credentials.session)

        super().__init__(report_dir, timestamp,
                                          services, skipped_services, result_format)

    def get_report_name(self):
        """
        Returns the name of the report using the provider's configuration
        """
        if self.profile:
            return f'aws-{self.profile}'
        elif self.account_id:
            return f'aws-{self.account_id}'
        else:
            return 'aws'

    def preprocessing(self, ip_ranges=None, ip_ranges_name_key=None):
        """
        Tweak the AWS config to match cross-service resources and clean any fetching artifacts

        :param ip_ranges:
        :param ip_ranges_name_key:
        :return: None
        """
        ip_ranges = [] if ip_ranges is None else ip_ranges

        # Various data processing calls
        # Note that order of processing can matter

        # TODO - this should be moved to the `finalize` method of the base resource, as it's not cross-service
        self._map_all_subnets()

        # TODO - this should be moved to the `finalize` method of the base resource, as it's not cross-service
        if 'ec2' in self.service_list:
            self._map_all_sgs()
            self._add_security_group_name_to_ec2_grants()
            self._check_ec2_zone_distribution()
            self._add_last_snapshot_date_to_ec2_volumes()

        if 'ec2' in self.service_list and 'iam' in self.service_list:
            self._match_instances_and_roles()
        
        if 'ec2' in self.service_list and 'vpc' in self.service_list:
            self._match_instances_and_vpcs()
            self._match_instances_and_subnets()
        
        if 'ec2' in self.service_list and 'codebuild' in self.service_list:
            self._update_sg_usage_codebuild()

        if 'awslambda' in self.service_list and 'iam' in self.service_list:
            self._match_lambdas_and_roles()

        if 'elbv2' in self.service_list and 'ec2' in self.service_list:
            self._add_security_group_data_to_elbv2()

        if 's3' in self.service_list and 'iam' in self.service_list:
            self._match_iam_policies_and_buckets()

        # TODO - this should be moved to the `finalize` method of the base resource, as it's not cross-service
        if 'elb' in self.services:
            self._parse_elb_policies()

        if 'emr' in self.service_list and 'ec2' in self.service_list and 'vpc' in self.service_list:
            self._set_emr_vpc_ids()

        self._add_cidr_display_name(ip_ranges, ip_ranges_name_key)

        super().preprocessing()

    def _add_cidr_display_name(self, ip_ranges, ip_ranges_name_key):
        if len(ip_ranges):
            callback_args = {'ip_ranges': ip_ranges,
                             'ip_ranges_name_key': ip_ranges_name_key}
            self._go_to_and_do(self.services['ec2'],
                               ['regions', 'vpcs', 'security_groups', 'rules', 'protocols', 'ports'],
                               ['services', 'ec2'],
                               put_cidr_name,
                               callback_args)

    def _add_security_group_name_to_ec2_grants(self):
        """
        Github issue #24: display the security group names in the list of grants (added here to have ligher JS code)
        """
        self._go_to_and_do(self.services['ec2'],
                           ['regions', 'vpcs', 'security_groups', 'rules', 'protocols', 'ports', 'security_groups'],
                           [],
                           self.add_security_group_name_to_ec2_grants_callback,
                           {'AWSAccountId': self.account_id})

    def _add_security_group_data_to_elbv2(self):
        def check_security_group_rules(lb, index, traffic_type):
            none = 'N/A'
            if traffic_type == 'ingress':
                output = 'valid_inbound_rules'
            elif traffic_type == 'egress':
                output = 'valid_outbound_rules'
            for protocol in lb['security_groups'][index]['rules'][traffic_type]['protocols']:
                for port in lb['security_groups'][index]['rules'][traffic_type]['protocols'][protocol]['ports']:
                    lb['security_groups'][index][output] = True
                    if port not in lb['listeners'] and port != none:
                        lb['security_groups'][index][output] = False

        ec2_config = self.services['ec2']
        elbv2_config = self.services['elbv2']
        for region in elbv2_config['regions']:
            for vpc in elbv2_config['regions'][region]['vpcs']:
                for lb in elbv2_config['regions'][region]['vpcs'][vpc]['lbs']:
                    for i in range(0, len(elbv2_config['regions'][region]['vpcs'][vpc]['lbs'][lb]['security_groups'])):
                        for sg in ec2_config['regions'][region]['vpcs'][vpc]['security_groups']:
                            group_id = elbv2_config['regions'][region]['vpcs'][vpc]['lbs'][lb]['security_groups'][i][
                                'GroupId']
                            if 'GroupId' in elbv2_config['regions'][region]['vpcs'][vpc]['lbs'][lb]['security_groups'][
                                i] and group_id == sg:
                                elbv2_config['regions'][region]['vpcs'][vpc]['lbs'][lb]['security_groups'][i] = \
                                    ec2_config['regions'][region]['vpcs'][vpc]['security_groups'][sg]
                                elbv2_config['regions'][region]['vpcs'][vpc]['lbs'][lb]['security_groups'][i][
                                    'GroupId'] = group_id

                        check_security_group_rules(
                            elbv2_config['regions'][region]['vpcs'][vpc]['lbs'][lb], i, 'ingress')
                        check_security_group_rules(
                            elbv2_config['regions'][region]['vpcs'][vpc]['lbs'][lb], i, 'egress')

    def _check_ec2_zone_distribution(self):
        regions = self.services['ec2']['regions'].values()
        self.services['ec2']['number_of_regions_with_instances'] = sum(r['instances_count'] > 0 for r in regions)

        for regions in self.services['ec2']['regions'].values():
            instances_availability_zones = set()
            for vpcs in regions['vpcs'].values():
                for instance in vpcs['instances'].values():
                    instances_availability_zones.add(instance.get('availability_zone'))
            regions['instances_availability_zones'] = len(instances_availability_zones)

    def _add_last_snapshot_date_to_ec2_volumes(self):
        for region in self.services['ec2']['regions'].values():
            for volumeId, volume in region.get('volumes').items():
                completed_snapshots = [s for s in region['snapshots'].values() if
                                       s['volume_id'] == volumeId and s['state'] == 'completed']
                sorted_snapshots = sorted(
                    completed_snapshots, key=lambda s: s['start_time'], reverse=True)
                volume['LastSnapshotDate'] = sorted_snapshots[0]['start_time'] if len(
                    sorted_snapshots) > 0 else None

    def add_security_group_name_to_ec2_grants_callback(self, current_config, path, current_path, ec2_grant,
                                                       callback_args):
        sg_id = ec2_grant['GroupId']
        if sg_id in current_path:
            target = current_path[:(current_path.index(sg_id) + 1)]
            ec2_grant['GroupName'] = get_value_at(self.services['ec2'], target, 'name')
        elif 'UserId' in ec2_grant and ec2_grant['UserId'] == callback_args['AWSAccountId']:
            if 'VpcId' in ec2_grant:
                target = current_path[:(current_path.index('vpcs') + 1)]
                target.append(ec2_grant['VpcId'])
                target.append('security_groups')
                target.append(sg_id)
            else:
                target = current_path[:(
                        current_path.index('security_groups') + 1)]
                target.append(sg_id)
            ec2_grant['GroupName'] = get_value_at(self.services['ec2'], target, 'name')
        elif 'PeeringStatus' in ec2_grant:
            # Can't infer the name of the SG in the peered account
            pass
        else:
            print_warning('Failed to handle EC2 grant: %s' % ec2_grant)

    def process_network_acls_callback(self, current_config, path, current_path, privateip_id, callback_args):
        # Check if the network ACL allows all traffic from all IP addresses
        self._process_network_acls_check_for_allow_all(
            current_config, 'ingress')
        self._process_network_acls_check_for_allow_all(
            current_config, 'egress')
        # Check if the network ACL only has the default rules
        self._process_network_acls_check_for_aws_default(
            current_config, 'ingress')
        self._process_network_acls_check_for_aws_default(
            current_config, 'egress')

    @staticmethod
    def _process_network_acls_check_for_allow_all(network_acl, direction):
        network_acl['allow_all_%s_traffic' % direction] = 0
        for rule_number in network_acl['rules'][direction]:
            rule = network_acl['rules'][direction][rule_number]
            if rule['RuleAction'] == 'deny':
                # If a deny rule appears before an allow all, do not raise the flag
                break
            if (rule['CidrBlock'] == '0.0.0.0/0') and (rule['RuleAction'] == 'allow') and (
                    rule['port_range'] == '1-65535') and (rule['protocol'] == 'ALL'):
                network_acl['allow_all_%s_traffic' % direction] = rule_number
                break

    @staticmethod
    def _process_network_acls_check_for_aws_default(network_acl, direction):
        if len(network_acl['rules'][direction]) == 2 and int(
                network_acl['allow_all_%s_traffic' % direction]) > 0 and '100' in network_acl['rules'][direction]:
            # Assume it is AWS' default rules because there are 2 rules (100 and 65535) and the first rule allows all
            # traffic
            network_acl['use_default_%s_rules' % direction] = True
        else:
            network_acl['use_default_%s_rules' % direction] = False

    def list_ec2_network_attack_surface_callback(self, current_config, path, current_path, privateip_id, callback_args):
        try:
            manage_dictionary(self.services['ec2'], 'external_attack_surface', {})
            if 'Association' in current_config and current_config['Association']:
                public_ip = current_config['Association']['PublicIp']
                self._security_group_to_attack_surface(self.services['ec2']['external_attack_surface'],
                                                       public_ip, current_path,
                                                       [g['GroupId']
                                                        for g in current_config['Groups']],
                                                       [])
                self._complete_information_on_ec2_attack_surface(current_config, current_path, public_ip)

            # IPv6
            if 'Ipv6Addresses' in current_config and len(current_config['Ipv6Addresses']) > 0:
                for ipv6 in current_config.get('Ipv6Addresses', []):
                    ip = ipv6['Ipv6Address']
                    self._security_group_to_attack_surface(self.services['ec2']['external_attack_surface'],
                                                           ip, current_path,
                                                           [g['GroupId'] for g in current_config['Groups']], [])
                    self._complete_information_on_ec2_attack_surface(current_config, current_path, ip)
        except Exception as e:
            print_exception(f"Error listing EC2 network attack surface: {e}")

    def _complete_information_on_ec2_attack_surface(self, current_config, current_path, public_ip):
        try:
            # Get the EC2 instance info
            ec2_info = self.services
            for p in current_path[1:-3]:
                ec2_info = ec2_info[p]
            # Fill the rest of the attack surface details on that IP
            self.services['ec2']['external_attack_surface'][public_ip]['InstanceName'] = ec2_info.get('name')
            if current_config is not None and 'PublicDnsName' in current_config.get('Association', {}):
                self.services['ec2']['external_attack_surface'][public_ip]['PublicDnsName'] = \
                    current_config.get('Association', {}).get('PublicDnsName')
        except Exception as e:
            print_exception(f"Error completing EC2 network attack surface information: {e}")

    def _map_all_sgs(self):
        sg_map = dict()
        self._go_to_and_do(self.services['ec2'],
                           ['regions', 'vpcs', 'security_groups'],
                           ['services', 'ec2'],
                           self.map_resource,
                           {'map': sg_map})
        self.sg_map = sg_map

    def _map_all_subnets(self):
        subnet_map = dict()
        self._go_to_and_do(self.services['vpc'],
                           ['regions', 'vpcs', 'subnets'],
                           ['services', 'vpc'],
                           self.map_resource,
                           {'map': subnet_map})
        self.subnet_map = subnet_map

    @staticmethod
    def map_resource(current_config, path, current_path, resource_id, callback_args):
        if resource_id not in callback_args['map']:
            callback_args['map'][resource_id] = {'region': current_path[3]}
            if len(current_path) > 5:
                callback_args['map'][resource_id]['vpc_id'] = current_path[5]

    def _match_iam_policies_and_buckets(self):
        s3_info = self.services['s3']
        iam_info = self.services['iam']
        if 'Action' in iam_info['permissions']:
            for action in (x for x in iam_info['permissions']['Action'] if
                           ((x.startswith('s3:') and x != 's3:ListAllMyBuckets') or (x == '*'))):
                for iam_entity in iam_info['permissions']['Action'][action]:
                    if 'Allow' in iam_info['permissions']['Action'][action][iam_entity]:
                        for allowed_iam_entity in iam_info['permissions']['Action'][action][iam_entity]['Allow']:
                            # For resource statements, we can easily rely on the existing permissions structure
                            if 'Resource' in \
                                    iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]:
                                for full_path in (x for x in
                                                  iam_info['permissions']['Action'][action][iam_entity]['Allow'][
                                                      allowed_iam_entity]['Resource'] if
                                                  x.startswith('arn:aws:s3:') or x == '*'):
                                    parts = full_path.split('/')
                                    bucket_name = parts[0].split(':')[-1]
                                    self._update_iam_permissions(s3_info, bucket_name, iam_entity, allowed_iam_entity,
                                                                 iam_info['permissions']['Action'][action][iam_entity][
                                                                     'Allow'][allowed_iam_entity]['Resource'][
                                                                     full_path])
                            # For notresource statements, we must fetch the policy document to determine which
                            # buckets are not protected
                            if 'NotResource' in iam_info['permissions']['Action'][action][iam_entity]['Allow'][
                                allowed_iam_entity]:
                                for full_path in (x for x in
                                                  iam_info['permissions']['Action'][action][iam_entity]['Allow'][
                                                      allowed_iam_entity]['NotResource'] if
                                                  x.startswith('arn:aws:s3:') or x == '*'):
                                    for policy_type in ['InlinePolicies', 'ManagedPolicies']:
                                        if policy_type in \
                                                iam_info['permissions']['Action'][action][iam_entity]['Allow'][
                                                    allowed_iam_entity]['NotResource'][full_path]:
                                            for policy in \
                                                    iam_info['permissions']['Action'][action][iam_entity]['Allow'][
                                                        allowed_iam_entity]['NotResource'][full_path][policy_type]:
                                                self._update_bucket_permissions(s3_info, iam_info, action, iam_entity,
                                                                                allowed_iam_entity, full_path,
                                                                                policy_type,
                                                                                policy)

    def _update_bucket_permissions(self, s3_info, iam_info, action, iam_entity, allowed_iam_entity, full_path,
                                   policy_type,
                                   policy_name):
        global policy
        allowed_buckets = []
        # By default, all buckets are allowed
        for bucket_name in s3_info['buckets']:
            allowed_buckets.append(bucket_name)
        if policy_type == 'InlinePolicies':
            policy = iam_info[iam_entity.title(
            )][allowed_iam_entity]['Policies'][policy_name]['PolicyDocument']
        elif policy_type == 'ManagedPolicies':
            policy = iam_info['ManagedPolicies'][policy_name]['PolicyDocument']
        else:
            print_error('Error, found unknown policy type.')
        for statement in policy['Statement']:
            for target_path in statement['NotResource']:
                parts = target_path.split('/')
                bucket_name = parts[0].split(':')[-1]
                path = '/' + '/'.join(parts[1:]) if len(parts) > 1 else '/'
                if (path == '/' or path == '/*') and (bucket_name in allowed_buckets):
                    # Remove bucket from list
                    allowed_buckets.remove(bucket_name)
                elif bucket_name == '*':
                    allowed_buckets = []
        policy_info = {policy_type: {}}
        policy_info[policy_type][policy_name] = \
            iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'][
                full_path][
                policy_type][policy_name]
        for bucket_name in allowed_buckets:
            self._update_iam_permissions(
                s3_info, bucket_name, iam_entity, allowed_iam_entity, policy_info)

    def _update_iam_permissions(self, s3_info, bucket_name, iam_entity, allowed_iam_entity, policy_info):
        if 's3' in self.service_list and 'iam' in self.service_list:  # validate both services were included in run
            if bucket_name != '*' and bucket_name in s3_info['buckets']:
                bucket = s3_info['buckets'][bucket_name]
                manage_dictionary(bucket, iam_entity, {})
                manage_dictionary(bucket, iam_entity + '_count', 0)
                if allowed_iam_entity not in bucket[iam_entity]:
                    bucket[iam_entity][allowed_iam_entity] = {}
                    bucket[iam_entity + '_count'] = bucket[iam_entity + '_count'] + 1

                if 'inline_policies' in policy_info:
                    manage_dictionary(
                        bucket[iam_entity][allowed_iam_entity], 'inline_policies', {})
                    bucket[iam_entity][allowed_iam_entity]['inline_policies'].update(
                        policy_info['inline_policies'])
                if 'policies' in policy_info:
                    manage_dictionary(bucket[iam_entity]
                                      [allowed_iam_entity], 'policies', {})
                    bucket[iam_entity][allowed_iam_entity]['policies'].update(
                        policy_info['policies'])
            elif bucket_name == '*':
                for bucket in s3_info['buckets']:
                    self._update_iam_permissions(
                        s3_info, bucket, iam_entity, allowed_iam_entity, policy_info)
            else:
                # Could be an error or cross-account access, ignore
                pass

    def match_network_acls_and_subnets_callback(self, current_config, path, current_path, acl_id, callback_args):
        for association in current_config['Associations']:
            subnet_path = current_path[:-1] + \
                          ['subnets', association['SubnetId']]
            subnet = get_object_at(self, subnet_path)
            subnet['network_acl'] = acl_id

    def _match_instances_and_subnets(self):
        ec2_instances = self._get_ec2_instances_details(['id', 'vpc', 'region', 'SubnetId'])  # fetch all EC2 instances with only required fields
        for instance in ec2_instances.values():
            subnet = self.services['vpc']['regions'][instance['region']]['vpcs'][instance['vpc']]['subnets'][instance['SubnetId']]  # find the subnet reference
            manage_dictionary(subnet, 'instances', [])  # initialize instances list for the subnet (if not already set)
            if instance['id'] not in subnet['instances']:  # if instance is not already mapped to the subnet
                subnet['instances'].append(instance['id'])  # append EC2 instance ID to instance list in subnet

    def _get_ec2_instances_details(self, details=None):
        """
        Fetches a list of EC2 instances 

        :param details [str]:       (Optional) List of details to be included, if not specified, all details will be included
        :return:                    A dictionary of EC2 instances with the specified details
        """
        ec2_instances = {}
        for ec2_region_id, ec2_region_data in self.services['ec2']['regions'].items():
            if ec2_region_data['instances_count'] > 0:
                for region_vpc_id, region_vpc_data in ec2_region_data['vpcs'].items():
                    if region_vpc_data['instances_count'] > 0:
                        for ec2_instance_id, ec2_instance_data in region_vpc_data['instances'].items():
                            ec2_instances[ec2_instance_id] = ec2_instance_data.copy()
                            ec2_instances[ec2_instance_id]['region'] = ec2_region_id
                            ec2_instances[ec2_instance_id]['vpc'] = region_vpc_id
        if details is not None:
            for instance_key in ec2_instances.keys():
                for detail in list(ec2_instances[instance_key].keys()):
                    if detail not in details:
                        ec2_instances[instance_key].pop(detail, None)
        return ec2_instances

    def _match_instances_and_vpcs(self):
        ec2_instances = self._get_ec2_instances_details(['id', 'vpc', 'region'])  # fetch all EC2 instances with only required fields
        for instance in ec2_instances.values():
            vpc = self.services['vpc']['regions'][instance['region']]['vpcs'][instance['vpc']]  # find the VPC reference
            manage_dictionary(vpc, 'instances', [])  # initialize instances list for the VPC (if not already set)
            if instance['id'] not in vpc['instances']:  # if instance is not already mapped to the VPC
                vpc['instances'].append(instance['id'])  # append EC2 instance ID to instance list in VPC

    def _match_instances_and_roles(self):
        if 'ec2' in self.service_list and 'iam' in self.service_list:  # validate both services were included in run
            ec2_config = self.services['ec2']
            iam_config = self.services['iam']
            role_instances = {}
            for r in ec2_config['regions']:
                for v in ec2_config['regions'][r]['vpcs']:
                    if 'instances' in ec2_config['regions'][r]['vpcs'][v]:
                        for i in ec2_config['regions'][r]['vpcs'][v]['instances']:
                            instance_profile = ec2_config['regions'][r]['vpcs'][v]['instances'][i]['IamInstanceProfile']
                            instance_profile_id = instance_profile['Id'] if instance_profile else None
                            if instance_profile_id:
                                manage_dictionary(
                                    role_instances, instance_profile_id, [])
                                role_instances[instance_profile_id].append(i)
            for role_id in iam_config['roles']:
                iam_config['roles'][role_id]['instances_count'] = 0
                for instance_profile_id in iam_config['roles'][role_id]['instance_profiles']:
                    if instance_profile_id in role_instances:
                        iam_config['roles'][role_id]['instance_profiles'][instance_profile_id]['instances'] = \
                            role_instances[instance_profile_id]
                        iam_config['roles'][role_id]['instances_count'] += len(
                            role_instances[instance_profile_id])

    def _match_lambdas_and_roles(self):
        if self.services.get('awslambda') and self.services.get('iam'):
            awslambda_config = self.services['awslambda']
            iam_config = self.services['iam']
            awslambda_funtions = {}
            for r in awslambda_config['regions']:
                for lambda_function in awslambda_config['regions'][r]['functions']:
                    awslambda_function = awslambda_config['regions'][r]['functions'][lambda_function]
                    awslambda_function['region'] = r
                    if awslambda_function['role_arn'] in awslambda_funtions:
                        awslambda_funtions[awslambda_function['role_arn']][awslambda_function['name']] = awslambda_function
                    else:
                        awslambda_funtions[awslambda_function['role_arn']] = {awslambda_function['name']: awslambda_function}
            for role_id in iam_config['roles']:
                iam_config['roles'][role_id]['awslambdas_count'] = 0
                if iam_config['roles'][role_id]['arn'] in awslambda_funtions:
                    iam_config['roles'][role_id]['awslambdas'] = awslambda_funtions[iam_config['roles'][role_id]['arn']]
                    iam_config['roles'][role_id]['awslambdas_count'] = len(awslambda_funtions[iam_config['roles'][role_id]['arn']])

    def process_vpc_peering_connections_callback(self, current_config, path, current_path, pc_id, callback_args):

        # Create a list of peering connection IDs in each VPC
        info = 'AccepterVpcInfo' if current_config['AccepterVpcInfo'][
                                        'OwnerId'] == self.account_id else 'RequesterVpcInfo'
        region = current_path[current_path.index('regions') + 1]
        vpc_id = current_config[info]['VpcId']
        if vpc_id not in self.services['vpc']['regions'][region]['vpcs']:
            region = current_config['AccepterVpcInfo']['Region']

        # handle edge case where the region wasn't included in the execution
        if region in self.services['vpc']['regions']:
            target = self.services['vpc']['regions'][region]['vpcs'][vpc_id]
            manage_dictionary(target, 'peering_connections', [])
            if pc_id not in target['peering_connections']:
                target['peering_connections'].append(pc_id)

        # VPC information for the peer'd VPC
        current_config['peer_info'] = copy.deepcopy(
            current_config['AccepterVpcInfo' if info == 'RequesterVpcInfo' else 'RequesterVpcInfo'])
        if 'PeeringOptions' in current_config['peer_info']:
            current_config['peer_info'].pop('PeeringOptions')
        if hasattr(self, 'organization') and current_config['peer_info']['OwnerId'] in self.organization:
            current_config['peer_info']['name'] = self.organization[current_config['peer_info']['OwnerId']][
                'Name']
        else:
            current_config['peer_info']['name'] = current_config['peer_info']['OwnerId']

    def match_roles_and_cloudformation_stacks_callback(self,
                                                       current_config, path, current_path, stack_id, callback_args):
        try:
            if 'RoleARN' not in current_config:
                return
            role_arn = current_config.pop('RoleARN')
            current_config['iam_role'] = self._get_role_info('arn', role_arn)
        except Exception as e:
            print_exception(f'Unable to match roles and CloudFormation stacks: {e}')

    def match_roles_and_vpc_flowlogs_callback(self, current_config, path, current_path, flowlog_id, callback_args):
        if 'DeliverLogsPermissionArn' not in current_config:
            return
        delivery_role_arn = current_config.pop('DeliverLogsPermissionArn')
        current_config['delivery_role'] = self._get_role_info(
            'arn', delivery_role_arn)

    def _get_role_info(self, attribute_name, attribute_value):
        try:
            iam_role_info = {'name': None, 'id': None}
            for role_id in self.services['iam'].get('roles', []):
                if self.services['iam']['roles'][role_id][attribute_name] == attribute_value:
                    iam_role_info['name'] = self.services['iam']['roles'][role_id]['name']
                    iam_role_info['id'] = role_id
                    break
            return iam_role_info
        except Exception as e:
            print_exception(f'Unable to get role info for attribute {attribute_name} with value {attribute_value}: {e}')

    def match_security_groups_and_resources_callback(self, current_config, path, current_path, resource_id,
                                                     callback_args):
        if 'ec2' in self.service_list:  # validate that the service was included in run
            service = current_path[1]
            original_resource_path = combine_paths(
                copy.deepcopy(current_path), [resource_id])
            resource = get_object_at(self, original_resource_path)
            if 'resource_id_path' not in callback_args:
                resource_type = current_path[-1]
                resource_path = copy.deepcopy(current_path)
                resource_path.append(resource_id)
            else:
                resource_path = combine_paths(copy.deepcopy(
                    current_path), callback_args['resource_id_path'])
                resource_id = resource_path[-1]
                resource_type = resource_path[-2]
            if 'status_path' in callback_args:
                status_path = combine_paths(copy.deepcopy(
                    original_resource_path), callback_args['status_path'])
                obj = get_object_at(self, status_path)
                if obj:
                    resource_status = obj.replace('.', '_')
                else:
                    resource_status = obj
            else:
                resource_status = None
            unknown_vpc_id = True if current_path[4] != 'vpcs' else False
            # Issue 89 & 91 : can instances have no security group?
            try:
                try:
                    sg_attribute = get_object_at(
                        resource, callback_args['sg_list_attribute_name'])
                except Exception as e:
                    return
                if type(sg_attribute) != list:
                    sg_attribute = [sg_attribute]
                for resource_sg in sg_attribute:
                    if type(resource_sg) == dict:
                        sg_id = resource_sg[callback_args['sg_id_attribute_name']]
                    else:
                        sg_id = resource_sg
                    if unknown_vpc_id and sg_id:
                        vpc_id = self.sg_map[sg_id]['vpc_id']
                        sg_base_path = copy.deepcopy(current_path[0:4])
                        sg_base_path[1] = 'ec2'
                        sg_base_path = sg_base_path + \
                                       ['vpcs', vpc_id, 'security_groups']
                    else:
                        sg_base_path = copy.deepcopy(current_path[0:6])
                        sg_base_path[1] = 'ec2'
                        sg_base_path.append('security_groups')
                    sg_path = copy.deepcopy(sg_base_path)
                    sg_path.append(sg_id)
                    sg = get_object_at(self, sg_path)
                    # Add usage information
                    manage_dictionary(sg, 'used_by', {})
                    manage_dictionary(sg['used_by'], service, {})
                    manage_dictionary(sg['used_by'][service], 'resource_type', {})
                    manage_dictionary(sg['used_by'][service]['resource_type'], resource_type, {
                    } if resource_status else [])
                    if resource_status:
                        manage_dictionary(
                            sg['used_by'][service]['resource_type'][resource_type], resource_status, [])
                        if resource_id not in sg['used_by'][service]['resource_type'][resource_type][resource_status]:
                            sg['used_by'][service]['resource_type'][resource_type][resource_status].append(
                                {'id': resource_id, 'name': resource['name']})
                    else:
                        sg['used_by'][service]['resource_type'][resource_type].append(
                            {'id': resource_id, 'name': resource['name']})
            except Exception as e:
                if resource_type == 'elbs' and current_path[5] == ec2_classic:
                    pass
                elif not self.services['ec2']:  # service not included in run
                    pass
                elif not str(e):
                    print_exception(f'Failed to parse {resource_type}')
                else:
                    print_exception(f'Failed to parse {resource_type}: {e}')

    def _set_emr_vpc_ids(self):
        clear_list = []
        self._go_to_and_do(self.services['emr'],
                           ['regions', 'vpcs'],
                           ['services', 'emr'],
                           self.set_emr_vpc_ids_callback,
                           {'clear_list': clear_list})
        for region in clear_list:
            self.services['emr']['regions'][region]['vpcs'].pop('EMR-UNKNOWN-VPC')

    def set_emr_vpc_ids_callback(self, current_config, path, current_path, vpc_id, callback_args):
        if vpc_id != 'EMR-UNKNOWN-VPC':
            return
        region = current_path[3]
        vpc_id = sg_id = subnet_id = None
        pop_list = []
        for cluster_id in current_config['clusters']:
            cluster = current_config['clusters'][cluster_id]
            if 'EmrManagedMasterSecurityGroup' in cluster['Ec2InstanceAttributes']:
                sg_id = cluster['Ec2InstanceAttributes']['EmrManagedMasterSecurityGroup']
            elif 'RequestedEc2SubnetIds' in cluster['Ec2InstanceAttributes']:
                subnet_id = cluster['Ec2InstanceAttributes']['RequestedEc2SubnetIds']
            else:
                print_warning('Unable to determine VPC id for EMR cluster %s' % str(cluster_id))
                continue
            if sg_id in self.sg_map:
                vpc_id = self.sg_map[sg_id]['vpc_id']
                pop_list.append(cluster_id)
            else:
                sid_found = False
                if subnet_id:
                    for sid in subnet_id:
                        if sid in self.subnet_map:
                            vpc_id = self.subnet_map[sid]['vpc_id']
                            pop_list.append(cluster_id)
                            sid_found = True
                if not sid_found:
                    print_warning('Unable to determine VPC id for %s' % (str(subnet_id) if subnet_id else str(sg_id)))
                    continue
            if vpc_id:
                region_vpcs_config = get_object_at(self, current_path)
                manage_dictionary(region_vpcs_config, vpc_id, {'clusters': {}})
                region_vpcs_config[vpc_id]['clusters'][cluster_id] = cluster
        for cluster_id in pop_list:
            current_config['clusters'].pop(cluster_id)
        if len(current_config['clusters']) == 0:
            callback_args['clear_list'].append(region)

    def sort_vpc_flow_logs_callback(self, current_config, path, current_path, flow_log_id, callback_args):
        attached_resource = current_config['resource_id']
        if attached_resource.startswith('vpc-'):
            vpc_path = combine_paths(
                current_path[0:4], ['vpcs', attached_resource])
            try:
                attached_vpc = get_object_at(self, vpc_path)
            except Exception:
                print_debug(
                    'It appears that the flow log %s is attached to a resource that was previously deleted (%s).' % (
                        flow_log_id, attached_resource))
                return
            manage_dictionary(attached_vpc, 'flow_logs', [])
            if flow_log_id not in attached_vpc['flow_logs']:
                attached_vpc['flow_logs'].append(flow_log_id)
            for subnet_id in attached_vpc['subnets']:
                manage_dictionary(
                    attached_vpc['subnets'][subnet_id], 'flow_logs', [])
                if flow_log_id not in attached_vpc['subnets'][subnet_id]['flow_logs']:
                    attached_vpc['subnets'][subnet_id]['flow_logs'].append(
                        flow_log_id)
        elif attached_resource.startswith('subnet-'):
            subnet_path = combine_paths(current_path[0:4],
                                        ['vpcs', self.subnet_map[attached_resource]['vpc_id'], 'subnets',
                                         attached_resource])
            subnet = get_object_at(self, subnet_path)
            manage_dictionary(subnet, 'flow_logs', [])
            if flow_log_id not in subnet['flow_logs']:
                subnet['flow_logs'].append(flow_log_id)
        else:
            print_warning('Resource %s attached to flow logs is not handled' % attached_resource)

    def get_db_attack_surface(self, current_config, path, current_path, db_id, callback_args):
        service = current_path[1]
        service_config = self.services[service]
        manage_dictionary(service_config, 'external_attack_surface', {})
        if (service == 'redshift' or service == 'rds') and 'PubliclyAccessible' in current_config and current_config[
            'PubliclyAccessible']:
            public_dns = current_config['Endpoint']['Address']
            listeners = [current_config['Endpoint']['Port']]
            security_groups = current_config['VpcSecurityGroups']
            self._security_group_to_attack_surface(service_config['external_attack_surface'], public_dns,
                                                   current_path, [
                                                       g['VpcSecurityGroupId'] for g in security_groups],
                                                   listeners)
        elif 'ConfigurationEndpoint' in current_config:
            # TODO : get the proper addresss
            public_dns = current_config['ConfigurationEndpoint']['Address'].replace(
                '.cfg', '')
            listeners = [current_config['ConfigurationEndpoint']['Port']]
            security_groups = current_config.get('SecurityGroups', {})
            self._security_group_to_attack_surface(service_config['external_attack_surface'], public_dns,
                                                   current_path, [
                                                       g['SecurityGroupId'] for g in security_groups],
                                                   listeners)
            # TODO :: Get Redis endpoint information

    def get_lb_attack_surface(self, current_config, path, current_path, elb_id, callback_args):
        try:
            public_dns = current_config['DNSName']
            elb_config = self.services[current_path[1]]
            manage_dictionary(elb_config, 'external_attack_surface', {})
            if current_path[1] == 'elbv2' and current_config['Type'] == 'network':
                # Network LBs do not have a security group, lookup listeners instead
                manage_dictionary(
                    elb_config['external_attack_surface'], public_dns, {'protocols': {}})
                for listener in current_config['listeners']:
                    protocol = current_config['listeners'][listener]['Protocol']
                    manage_dictionary(elb_config['external_attack_surface'][public_dns]['protocols'], protocol,
                                      {'ports': {}})
                    manage_dictionary(elb_config['external_attack_surface'][public_dns]['protocols'][protocol]['ports'],
                                      listener, {'cidrs': []})
                    elb_config['external_attack_surface'][public_dns]['protocols'][protocol]['ports'][listener][
                        'cidrs'].append({'CIDR': '0.0.0.0/0'})
            elif current_path[1] == 'elbv2' and current_config['Scheme'] == 'internet-facing':
                elb_config['external_attack_surface'][public_dns] = {
                    'protocols': {}}
                security_groups = [g['GroupId']
                                   for g in current_config['security_groups']]
                listeners = []
                for listener in current_config['listeners']:
                    listeners.append(listener)
                self._security_group_to_attack_surface(elb_config['external_attack_surface'], public_dns,
                                                       current_path, security_groups, listeners)
            elif current_config['Scheme'] == 'internet-facing':
                # Classic ELbs do not have a security group, lookup listeners instead
                public_dns = current_config['DNSName']
                manage_dictionary(elb_config['external_attack_surface'], public_dns, {
                    'protocols': {'TCP': {'ports': {}}}})
                for listener in current_config['listeners']:
                    manage_dictionary(elb_config['external_attack_surface'][public_dns]['protocols']['TCP']['ports'],
                                      listener, {'cidrs': []})
                    elb_config['external_attack_surface'][public_dns]['protocols']['TCP']['ports'][listener][
                        'cidrs'].append({'CIDR': '0.0.0.0/0'})
        except Exception as e:
            print_exception(f'Failed to get LB attack surface: {e}')


    def _security_group_to_attack_surface(self, attack_surface_config, public_ip, current_path,
                                          security_groups, listeners=None):
        try:
            listeners = [] if listeners is None else listeners
            manage_dictionary(attack_surface_config, public_ip, {'protocols': {}})
            instance_path = current_path[:-3]
            if 'ec2' in self.service_list:  # validate that the service was included in run
                for sg_id in security_groups:
                    sg_path = copy.deepcopy(current_path[0:6])
                    sg_path[1] = 'ec2'
                    sg_path.append('security_groups')
                    sg_path.append(sg_id)
                    sg_path.append('rules')
                    sg_path.append('ingress')
                    ingress_rules = get_object_at(self, sg_path)
                    for p in ingress_rules['protocols']:
                        for port in ingress_rules['protocols'][p]['ports']:
                            if len(listeners) == 0 and 'cidrs' in ingress_rules['protocols'][p]['ports'][port]:
                                manage_dictionary(
                                    attack_surface_config[public_ip]['protocols'], p, {'ports': {}})
                                manage_dictionary(attack_surface_config[public_ip]['protocols'][p]['ports'], port,
                                                  {'cidrs': []})
                                attack_surface_config[public_ip]['protocols'][p]['ports'][port]['cidrs'] += \
                                    ingress_rules['protocols'][p]['ports'][port]['cidrs']
                            else:
                                ports = port.split('-')
                                if len(ports) > 1:
                                    try:
                                        if port[0]:
                                            port_min = int(ports[0])
                                        else:
                                            port_min = None
                                        if port[1]:
                                            port_max = int(ports[1])
                                        else:
                                            port_max = None
                                    except Exception as e:
                                        port_min = None
                                        port_max = None
                                elif port == 'N/A':
                                    port_min = port_max = None
                                elif port == 'ALL':
                                    port_min = 0
                                    port_max = 65535
                                elif p == 'ICMP':
                                    port_min = port_max = None
                                else:
                                    port_min = port_max = int(port)
                                for listener in listeners:
                                    if (port_min and port_max) and port_min < int(listener) < port_max and \
                                            'cidrs' in ingress_rules['protocols'][p]['ports'][port]:
                                        manage_dictionary(
                                            attack_surface_config[public_ip]['protocols'], p, {'ports': {}})
                                        manage_dictionary(attack_surface_config[public_ip]['protocols'][p]['ports'],
                                                          str(listener), {'cidrs': []})
                                        attack_surface_config[public_ip]['protocols'][p]['ports'][str(listener)]['cidrs'] += \
                                            ingress_rules['protocols'][p]['ports'][port]['cidrs']
        except Exception as e:
            print_exception(f'Failed to match SG to attack surface: {e}')

    def _parse_elb_policies(self):
        self._go_to_and_do(self.services['elb'],
                           ['regions'],
                           [],
                           self.parse_elb_policies_callback,
                           {})

    def parse_elb_policies_callback(self, current_config, path, current_path, region_id, callback_args):
        region_config = get_object_at(self, ['services', 'elb', ] + current_path + [region_id])
        region_config['elb_policies'] = current_config['elb_policies']
        for policy_id in region_config['elb_policies']:
            if region_config['elb_policies'][policy_id]['PolicyTypeName'] != 'SSLNegotiationPolicyType':
                continue
            # protocols, options, ciphers
            policy = region_config['elb_policies'][policy_id]
            protocols = {}
            options = {}
            ciphers = {}
            for attribute in policy['PolicyAttributeDescriptions']:
                if attribute['AttributeName'] in ['Protocol-SSLv3', 'Protocol-TLSv1', 'Protocol-TLSv1.1',
                                                  'Protocol-TLSv1.2']:
                    protocols[attribute['AttributeName']] = attribute['AttributeValue']
                elif attribute['AttributeName'] in ['Server-Defined-Cipher-Order']:
                    options[attribute['AttributeName']] = attribute['AttributeValue']
                elif attribute['AttributeName'] == 'Reference-Security-Policy':
                    policy['reference_security_policy'] = attribute['AttributeValue']
                else:
                    ciphers[attribute['AttributeName']] = attribute['AttributeValue']
                policy['protocols'] = protocols
                policy['options'] = options
                policy['ciphers'] = ciphers

    def _update_sg_usage_codebuild(self):
        try:
            for region in self.services['codebuild']['regions']:
                for codebuild_project in self.services['codebuild']['regions'][region]['build_projects']:
                    if 'vpc' in self.services['codebuild']['regions'][region]['build_projects'][codebuild_project] and 'security_groups' in self.services['codebuild']['regions'][region]['build_projects'][codebuild_project]:
                        cb_project = self.services['codebuild']['regions'][region]['build_projects'][codebuild_project]
                        for cb_project_sg in cb_project['security_groups']:
                            manage_dictionary(self.services['ec2']['regions'][region]['vpcs'][cb_project['vpc']]['security_groups'][cb_project_sg], 'used_by', {'resource_type': {'codebuild_project': []}})
                            self.services['ec2']['regions'][region]['vpcs'][cb_project['vpc']]['security_groups'][cb_project_sg]['used_by']['resource_type']['codebuild_project'].append({
                                'id': cb_project['arn'], 'name': cb_project['name']
                            })
        except Exception as e:
            print_exception(f'Failed to update security group usage for CodeBuild: {e}')



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class DiagnosticSettings(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        diagnostic_dict = {}
        diagnostic_dict['diagnostic_exist'] = await self.facade.loggingmonitoring.get_subscription_diagnostic_settings(
            self.subscription_id) != []
        self[self.subscription_id] = diagnostic_dict



import json
import logging
from getpass import getpass

import requests
from ScoutSuite.core.console import print_exception

from azure.identity import UsernamePasswordCredential, AzureCliCredential, ClientSecretCredential, \
    ManagedIdentityCredential, DeviceCodeCredential
from ScoutSuite.providers.base.authentication_strategy import AuthenticationStrategy, AuthenticationException

AUTHORITY_HOST_URI = 'https://login.microsoftonline.com/'
AZURE_CLI_CLIENT_ID = "04b07795-8ddb-461a-bbee-02f9e1bf7b46"


class AzureCredentials:

    def __init__(self,
                 identity_credentials,
                 tenant_id=None, default_subscription_id=None,
                 context=None):

        self.identity_credentials = identity_credentials  # Azure Resource Manager API credentials
        self.tenant_id = tenant_id
        self.default_subscription_id = default_subscription_id
        self.context = context

    def get_tenant_id(self):
        if self.tenant_id:
            return self.tenant_id
        elif hasattr(self.identity_credentials, 'tenant_id'):
            return self.identity_credentials['tenant_id']

        else:
            # Additional request for CLI & MSI authentication
            try:
                access_token = self.identity_credentials.get_token("https://management.core.windows.net/.default")
                h = {'Authorization': f'Bearer {access_token.token}'}
                r = requests.get('https://management.azure.com/tenants?api-version=2020-01-01', headers=h)
                r2 = r.json()
                return r2.get('value')[0].get('tenantId')
            except Exception as e:
                print_exception(f'Unable to infer tenant ID: {e}')
                return None

    def get_credentials(self):
        return self.identity_credentials


class AzureAuthenticationStrategy(AuthenticationStrategy):

    def authenticate(self,
                     cli=None, user_account=None, user_account_browser=None,
                     service_principal=None, file_auth=None, msi=None,
                     tenant_id=None,
                     subscription_id=None,
                     client_id=None, client_secret=None,
                     username=None, password=None,
                     programmatic_execution=False,
                     **kargs):
        """
        Implements authentication for the Azure provider
        """
        try:

            # Set logging level to error for libraries as otherwise generates a lot of warnings
            logging.getLogger('azure.identity').setLevel(logging.ERROR)
            logging.getLogger('azure.core.pipeline').setLevel(logging.ERROR)

            context = None

            if cli:
                identity_credentials = AzureCliCredential()

            elif user_account:

                if not (username and password):
                    if not programmatic_execution:
                        username = username if username else input("Username: ")
                        password = password if password else getpass("Password: ")
                    else:
                        raise AuthenticationException('Username or password not set')

                identity_credentials = UsernamePasswordCredential(AZURE_CLI_CLIENT_ID, username, password,
                                                                  authority=AUTHORITY_HOST_URI, tenant_id=tenant_id)

            elif user_account_browser:

                identity_credentials = DeviceCodeCredential(authority=AUTHORITY_HOST_URI,tenant_id=tenant_id,client_id=AZURE_CLI_CLIENT_ID)

            elif service_principal:

                if not tenant_id:
                    if not programmatic_execution:
                        tenant_id = input("Tenant ID: ")
                    else:
                        raise AuthenticationException('No Tenant ID set')

                if not client_id:
                    if not programmatic_execution:
                        client_id = input("Client ID: ")
                    else:
                        raise AuthenticationException('No Client ID set')

                if not client_secret:
                    if not programmatic_execution:
                        client_secret = getpass("Client secret: ")
                    else:
                        raise AuthenticationException('No Client Secret set')

                identity_credentials = ClientSecretCredential(
                    client_id=client_id,
                    client_secret=client_secret,
                    tenant_id=tenant_id
                )

            elif file_auth:

                data = json.loads(file_auth.read())
                tenant_id = data.get('tenantId')
                client_id = data.get('clientId')
                client_secret = data.get('clientSecret')

                identity_credentials = ClientSecretCredential(
                    client_id=client_id,
                    client_secret=client_secret,
                    tenant_id=tenant_id
                )

            elif msi:
                identity_credentials = ManagedIdentityCredential()

            else:
                raise AuthenticationException('Unknown authentication method')

            # Getting token to authenticate and detect AuthenticationException
            identity_credentials.get_token("https://management.core.windows.net/.default")

            return AzureCredentials(
                identity_credentials,
                tenant_id, subscription_id,
                context)

        except Exception as e:
            if 'Authentication failed: Unable to find wstrust endpoint from MEX. This typically happens when ' \
               'attempting MSA accounts. More details available here. ' \
               'https://github.com/AzureAD/microsoft-authentication-library-for-python/' \
               'wiki/Username-Password-Authentication' in e.args:

                raise AuthenticationException(
                    'You are likely authenticating with a Microsoft Account. '
                    'This authentication mode only support Azure Active Directory principal authentication.')
            raise AuthenticationException(e)



from azure.mgmt.rdbms.mysql import MySQLManagementClient
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.core.console import print_exception
from ScoutSuite.utils import get_user_agent


class MySQLDatabaseFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = MySQLManagementClient(self.credentials.get_credentials(),
                                       subscription_id=subscription_id,
                                       user_agent=get_user_agent())
        return client

    async def get_servers(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.servers.list())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve mySQL servers: {e}')
            return []



from ScoutSuite.providers.gcp.resources.projects import Projects
from ScoutSuite.providers.gcp.resources.stackdriverlogging.logging_metrics import LoggingMetrics
from ScoutSuite.providers.gcp.resources.stackdriverlogging.sinks import Sinks
from ScoutSuite.providers.gcp.resources.stackdriverlogging.metrics import Metrics


class StackdriverLogging(Projects):
    _children = [ 
        (Sinks, 'sinks'),
        (Metrics, 'metrics'),
        (LoggingMetrics, 'logging_metrics')
    ]



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class Disks(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_disk in await self.facade.virtualmachines.get_disks(self.subscription_id):
            id, disk = self._parse_disk(raw_disk)
            self[id] = disk

    def _parse_disk(self, raw_disk):
        disk_dict = {}

        disk_dict['id'] = get_non_provider_id(raw_disk.id)
        disk_dict['unique_id'] = getattr(raw_disk, 'unique_id', None)
        disk_dict['name'] = raw_disk.name
        disk_dict['type'] = raw_disk.type
        disk_dict['location'] = raw_disk.location
        disk_dict['tags'] = raw_disk.tags
        disk_dict['managed_by'] = raw_disk.managed_by
        disk_dict['sku'] = raw_disk.sku
        disk_dict['zones'] = raw_disk.zones
        disk_dict['time_created'] = raw_disk.time_created
        disk_dict['os_type'] = raw_disk.os_type
        disk_dict['hyper_vgeneration'] = raw_disk.hyper_v_generation
        disk_dict['creation_data'] = raw_disk.creation_data
        disk_dict['disk_size_gb'] = raw_disk.disk_size_gb
        disk_dict['disk_size_bytes'] = getattr(raw_disk, 'disk_size_bytes', None)
        disk_dict['provisioning_state'] = raw_disk.provisioning_state
        disk_dict['disk_iops_read_write'] = raw_disk.disk_iops_read_write
        disk_dict['disk_mbps_read_write'] = raw_disk.disk_m_bps_read_write
        disk_dict['disk_state'] = raw_disk.disk_state
        disk_dict['additional_properties'] = raw_disk.additional_properties

        if hasattr(raw_disk, 'encryption'):
            disk_dict['encryption_type'] = getattr(raw_disk.encryption, 'type', None)
        else:
            disk_dict['encryption_type'] = None

        if getattr(raw_disk, 'encryption_settings_collection', None):
            disk_dict['encryption_ade'] = raw_disk.encryption_settings_collection.enabled and \
                getattr(raw_disk.encryption_settings_collection, 'encryption_settings_version') in ['1.0','1.1']
        else:
            disk_dict['encryption_ade'] = False

        return disk_dict['id'], disk_dict





from ScoutSuite.providers.azure.authentication_strategy import AzureCredentials
from ScoutSuite.providers.azure.facade.aad import AADFacade
from ScoutSuite.providers.azure.facade.rbac import RBACFacade
from ScoutSuite.providers.azure.facade.keyvault import KeyVaultFacade
from ScoutSuite.providers.azure.facade.network import NetworkFacade
from ScoutSuite.providers.azure.facade.resourcemanagement import ResourceManagementFacade
from ScoutSuite.providers.azure.facade.securitycenter import SecurityCenterFacade
from ScoutSuite.providers.azure.facade.sqldatabase import SQLDatabaseFacade
from ScoutSuite.providers.azure.facade.storageaccounts import StorageAccountsFacade
from ScoutSuite.providers.azure.facade.virtualmachines import VirtualMachineFacade
from ScoutSuite.providers.azure.facade.appservice import AppServiceFacade
from ScoutSuite.providers.azure.facade.mysqldatabase import MySQLDatabaseFacade
from ScoutSuite.providers.azure.facade.postgresqldatabse import PostgreSQLDatabaseFacade
from ScoutSuite.providers.azure.facade.loggingmonitoring import LoggingMonitoringFacade

from azure.mgmt.resource import SubscriptionClient
from ScoutSuite.providers.base.authentication_strategy import AuthenticationException
from ScoutSuite.utils import get_user_agent

from ScoutSuite.core.console import print_info, print_exception

# Try to import proprietary services
try:
    from ScoutSuite.providers.azure.facade.appgateway_private import AppGatewayFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.azure.facade.loadbalancer_private import LoadBalancerFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.azure.facade.rediscache_private import RedisCacheFacade
except ImportError:
    pass


class AzureFacade:
    def __init__(self,
                 credentials: AzureCredentials,
                 subscription_ids=[], all_subscriptions=False,
                 programmatic_execution=False):

        self.credentials = credentials
        self.programmatic_execution = programmatic_execution

        self.subscription_list = []
        self.subscription_ids = subscription_ids
        self.all_subscriptions = all_subscriptions

        self.aad = AADFacade(credentials)
        self.rbac = RBACFacade(credentials)
        self.keyvault = KeyVaultFacade(credentials)
        self.virtualmachines = VirtualMachineFacade(credentials)
        self.network = NetworkFacade(credentials)
        self.securitycenter = SecurityCenterFacade(credentials)
        self.sqldatabase = SQLDatabaseFacade(credentials)
        self.storageaccounts = StorageAccountsFacade(credentials)
        self.appservice = AppServiceFacade(credentials)
        self.mysqldatabase = MySQLDatabaseFacade(credentials)
        self.postgresqldatabase = PostgreSQLDatabaseFacade(credentials)
        self.loggingmonitoring = LoggingMonitoringFacade(credentials)
        self.resourcemanagement = ResourceManagementFacade(credentials)

        # Instantiate facades for proprietary services
        try:
            self.appgateway = AppGatewayFacade(credentials)
        except NameError:
            pass
        try:
            self.loadbalancer = LoadBalancerFacade(credentials)
        except NameError:
            pass
        try:
            self.rediscache = RedisCacheFacade(credentials)
        except NameError:
            pass

        self._set_subscriptions()

    async def get_subscriptions(self):
        if self.subscription_list:
            return self.subscription_list
        else:
            self._set_subscriptions()

    def _set_subscriptions(self):

        # Create the client
        subscription_client = SubscriptionClient(self.credentials.get_credentials(), user_agent=get_user_agent())
        # Get all the accessible subscriptions
        accessible_subscriptions_list = list(subscription_client.subscriptions.list())

        if not accessible_subscriptions_list:
            raise AuthenticationException('The provided credentials do not have access to any subscriptions')

        # Final list, start empty
        subscriptions_list = []

        # No subscription provided, infer
        if not (self.subscription_ids or self.all_subscriptions):
            try:
                # Tries to read the subscription list
                print_info('No subscription set, inferring')
                s = next(subscription_client.subscriptions.list())
            except StopIteration:
                print_info('Unable to infer a subscription')
                # If the user cannot read subscription list, ask Subscription ID:
                if not self.programmatic_execution:
                    s = input('Subscription ID: ')
                else:
                    print_exception('Unable to infer a Subscription ID')
                    # raise
            finally:
                subscriptions_list.append(s)

        # All subscriptions
        elif self.all_subscriptions:
            subscriptions_list = accessible_subscriptions_list

        # A specific set of subscriptions
        elif self.subscription_ids:
            # Only include accessible subscriptions
            subscriptions_list = [s for s in accessible_subscriptions_list if
                                  s.subscription_id in self.subscription_ids]
            # Verbose skip
            for s in self.subscription_ids:
                if not any(subs.subscription_id == s for subs in accessible_subscriptions_list):
                    raise AuthenticationException('Subscription {} does not exist or is not accessible '
                                                  'with the provided credentials'.format(s))

        # Other == error
        else:
            raise AuthenticationException('Unknown Azure subscription option')

        if subscriptions_list and len(subscriptions_list) > 0:
            self.subscription_list = subscriptions_list
            if len(subscriptions_list) == 1:
                print_info('Running against subscription {}'.format(subscriptions_list[0].subscription_id))
            else:
                print_info('Running against {} subscriptions'.format(len(subscriptions_list)))
        else:
            raise AuthenticationException('No subscriptions to scan')



from ScoutSuite.core.console import print_error
from ScoutSuite.providers.kubernetes.facade.base import KubernetesBaseFacade


class ExtraFacade(KubernetesBaseFacade):
    def __init__(self, credentials):
        super().__init__(credentials)
        self.api_groups = None

    def get_resource_definitions(self) -> dict:
        if self.resource_definitions != None:
            return self.resource_definitions

        self.resource_definitions = self.get('/apis')

        return self.resource_definitions

    def get_resources(self) -> dict:
        if self.data != None: return self.data

        data = {}

        extra_resources = self.get_resource_definitions()
        for group in extra_resources['groups']:
            for version in group['versions']:
                endpoint = f'''/apis/{version['groupVersion']}'''

                api_resources = self.get(endpoint)
                if not api_resources:
                    continue

                for api_resource in api_resources['resources']:
                    if 'list' not in api_resource['verbs']: continue
                    endpoint = f'''/apis/{version['groupVersion']}/{api_resource['name']}'''

                    api_resources = self.get(endpoint)
                    if not api_resources:
                        continue

                    resources = self.get(endpoint)['items']
                    key = api_resource['kind']
                    data[key] = data.get(key, {})
                    data[key][version['groupVersion']] = resources

        self.data = self.parse_data(data)
        return self.data


# Import AWS utils
from ScoutSuite.providers.aws.utils import (
    get_keys,
    no_camel,
    get_name,
    is_throttled,
    get_aws_account_id,
    get_partition_name,
    snake_keys,
)
from ScoutSuite.utils import *
import collections
import unittest
from unittest import mock
import datetime

#
# Test methods for ScoutSuite/utils.py
#
class TestScoutUtilsClass(unittest.TestCase):
    def test_format_service_name(self):
        assert format_service_name("iAm") == "IAM"
        assert format_service_name("cloudformation") == "CloudFormation"

    def test_get_keys(self):
        test1 = {"a": "b", "c": "d"}
        test2 = {"a": "", "e": "f"}
        get_keys(test1, test2, "a")
        assert test2["a"] == "b"
        assert "c" not in test2
        get_keys(test1, test2, "c")
        assert test2["c"] == "d"

    def test_no_camel(self):
        assert no_camel("TestTest") == "test_test"

    def test_is_throttled(self):
        CustomException = collections.namedtuple("CustomException", "response")
        # test the throttling cases
        for t in ["Throttling", "RequestLimitExceeded", "ThrottlingException"]:
            e = CustomException(response={"Error": {"Code": t}})
            assert is_throttled(e)
        # test the non-throttling exception
        e = CustomException(response={"Error": {"Code": "Not Thro_ttling"}})
        assert not is_throttled(e)
        # test the except block
        e = CustomException(response={"Error": ""})
        assert not is_throttled(e)

    def test_get_name(self):
        src = {
            "Tags": [
                {"Key": "Not Name", "Value": "xyz"},
                {"Key": "Name", "Value": "abc"},
            ],
            "default_attribute": "default_value",
        }
        dst = {}
        default_attribute = "default_attribute"
        assert get_name(src, dst, default_attribute) == "abc"
        assert dst["name"] == "abc"

        src = {
            "Tags": [{"Key": "Not Name", "Value": "xyz"}],
            "default_attribute": "default_value",
        }
        dst = {}
        default_attribute = "default_attribute"
        assert get_name(src, dst, default_attribute) == "default_value"
        assert dst["name"] == "default_value"

    def test_get_identity(self):
        with mock.patch(
            "ScoutSuite.providers.aws.utils.get_caller_identity",
            return_value={"Arn": "a:b:c:d:e:f:"},
        ):
            assert get_aws_account_id("") == "e"

    def test_get_partition_name(self):
        with mock.patch(
            "ScoutSuite.providers.aws.utils.get_caller_identity",
            return_value={"Arn": "a:b:c:d:e:f:"},
        ):
            assert get_partition_name("") == "b"

    def test_snake_case(self):
        src = {
            "AttributeDefinitions": [
                {"AttributeName": "string", "AttributeType": "S"},
            ],
            "TableName": "string",
            "KeySchema": [{"AttributeName": "string", "KeyType": "HASH"},],
            "TableStatus": "CREATING",
            "CreationDateTime": datetime.datetime(2015, 1, 1, 1, 1, 1, 1, None),
            "ProvisionedThroughput": {
                "LastIncreaseDateTime": datetime.datetime(2015, 1, 1, 1, 1, 1, 1, None),
                "LastDecreaseDateTime": datetime.datetime(2015, 1, 1, 1, 1, 1, 1, None),
                "NumberOfDecreasesToday": 123,
                "ReadCapacityUnits": 123,
                "WriteCapacityUnits": 123,
            },
            "TableSizeBytes": 123,
            "AnotherArray": [
                "One",
                "Two",
                "AnotherThing",
            ]
        }
        dest = {
            "attribute_definitions": [
                {"attribute_name": "string", "attribute_type": "S"},
            ],
            "table_name": "string",
            "key_schema": [{"attribute_name": "string", "key_type": "HASH"}],
            "table_status": "CREATING",
            "creation_date_time": datetime.datetime(2015, 1, 1, 1, 1, 1, 1, None),
            "provisioned_throughput": {
                "last_increase_date_time": datetime.datetime(
                    2015, 1, 1, 1, 1, 1, 1, None
                ),
                "last_decrease_date_time": datetime.datetime(
                    2015, 1, 1, 1, 1, 1, 1, None
                ),
                "number_of_decreases_today": 123,
                "read_capacity_units": 123,
                "write_capacity_units": 123,
            },
            "table_size_bytes": 123,
            "another_array": ["One", "Two", "AnotherThing"]
        }
        d = snake_keys(src)
        self.maxDiff = None
        self.assertEquals(d, dest)



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Tables(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super(Tables, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_tables = await self.facade.dynamodb.get_tables(self.region)
        for raw_table in raw_tables:
            name, resource = self._parse_table(raw_table)
            self[name] = resource

    def _parse_table(self, raw_table):
        table_dict = {}
        table_dict['name'] = raw_table.get('TableName')
        table_dict['id'] = raw_table.get('TableId')
        table_dict['arn'] = raw_table.get('TableArn')
        table_dict['attribute_definitions'] = raw_table.get('AttributeDefinitions')
        table_dict['key_schema'] = raw_table.get('KeySchema')
        table_dict['table_status'] = raw_table.get('TableStatus')
        table_dict['creation_date_time'] = raw_table.get('CreationDateTime')
        table_dict['provisioned_throughput'] = raw_table.get('ProvisionedThroughput')
        table_dict['table_size_bytes'] = raw_table.get('TableSizeBytes')
        table_dict['item_count'] = raw_table.get('ItemCount')
        table_dict['backup_summaries'] = raw_table.get('BackupSummaries')
        table_dict['continuous_backups'] = raw_table.get('ContinuousBackups')
        table_dict['tags'] = raw_table.get('tags')

        table_dict['automatic_backups_enabled'] = \
            raw_table['ContinuousBackups']['PointInTimeRecoveryDescription']['PointInTimeRecoveryStatus'] == 'ENABLED' \
                if 'ContinuousBackups' in raw_table else None

        if "SSEDescription" in raw_table:
            table_dict["sse_enabled"] = True
        else:
            table_dict["sse_enabled"] = False

        return table_dict['id'], table_dict



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class Settings(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_settings in await self.facade.securitycenter.get_settings(self.subscription_id):
            id, settings = self._parse_settings(
                raw_settings)
            self[id] = settings

    def _parse_settings(self, settings):
        settings_dict = {}
        settings_dict['id'] = get_non_provider_id(settings.id.lower())
        settings_dict['name'] = settings.name
        settings_dict['kind'] = settings.kind
        settings_dict['enabled'] = settings.enabled
        return settings_dict['id'], settings_dict



import unittest
from ScoutSuite.output.html import *
from ScoutSuite.output.utils import *

#
# Test methods for ScoutSuite/output
#
class TestScoutOutput(unittest.TestCase):

    ########################################
    # html.py
    ########################################

    def test_html_report(self):
        test_html = HTMLReport(report_name='test')
        assert (test_html.report_name == 'test')
        assert ('json' in test_html.get_content_from_folder(templates_type='conditionals'))
        assert ('json' in test_html.get_content_from_file(filename='/json_format.html'))

    def test_get_filename(self):
        assert ('scoutsuite-report/report.html' in get_filename("REPORT"))
        assert ('scoutsuite-report/scoutsuite-results/scoutsuite_results.js' in get_filename("RESULTS"))
        assert ('scoutsuite-results/scoutsuite_results.js' in get_filename("RESULTS", relative_path=True))
        assert ('scoutsuite-report/scoutsuite-results/scoutsuite_exceptions.js' in get_filename("EXCEPTIONS"))
        assert ('scoutsuite-results/scoutsuite_exceptions.js' in get_filename("EXCEPTIONS", relative_path=True))
        assert ('scoutsuite-report/scoutsuite-results/scoutsuite_errors.json' in get_filename("ERRORS"))
        assert ('scoutsuite-results/scoutsuite_errors.json' in get_filename("ERRORS", relative_path=True))



from azure.mgmt.rdbms.postgresql import PostgreSQLManagementClient
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.core.console import print_exception
from ScoutSuite.utils import get_user_agent


class PostgreSQLDatabaseFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = PostgreSQLManagementClient(self.credentials.get_credentials(),
                                            subscription_id=subscription_id,
                                            user_agent=get_user_agent())
        return client

    async def get_servers(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.servers.list())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve postgresSQL servers: {e}')
            return []

    async def get_config(self, resource_group_name, server_name,
                         subscription_id: str, configuration_name: str):
        try:
            client = self.get_client(subscription_id)
            val = await run_concurrently(
                lambda: client.configurations.get(resource_group_name, server_name, configuration_name)
            )
            return val
        except Exception as e:
            print_exception(f'Failed to retrieve server configuration: {e}')
            return []

    async def get_firewall_rules(self, resource_group_name, server_name, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.firewall_rules.list_by_server(resource_group_name, server_name))
            )
        except Exception as e:
            print_exception(f'Failed to retrieve firewalls rules: {e}')
            return []



from ScoutSuite.providers.azure.resources.base import AzureResources


class Policies(AzureResources):
    async def fetch_all(self):
        raw_policy = await self.facade.aad.get_policies()
        id, policy = await self._parse_policy(raw_policy)
        self[id] = policy

    async def _parse_policy(self, raw_policy):
        policy_dict = {}
        policy_dict['id'] = raw_policy.get('id')
        policy_dict['name'] = raw_policy.get('displayName')
        policy_dict['allow_invites_from'] = raw_policy.get('allowInvitesFrom')
        policy_dict[
            'allowed_to_sign_up_email_based_subscription'] = raw_policy.get('allowedToSignUpEmailBasedSubscriptions')
        policy_dict['allowed_to_use_SSPR'] = raw_policy.get('allowedToUseSSPR')
        policy_dict['allow_email_verified_users_to_join_organization'
                    ] = raw_policy.get('allowEmailVerifiedUsersToJoinOrganization')
        policy_dict['allowed_to_create_apps'] = raw_policy.get('defaultUserRolePermissions', {}).get('allowedToCreateApps')
        policy_dict['allowed_to_create_security_groups'
                    ] = raw_policy.get('defaultUserRolePermissions', {}).get('allowedToCreateSecurityGroups')
        policy_dict[
            'allowed_to_read_other_users'] = raw_policy.get('defaultUserRolePermissions', {}).get('allowedToReadOtherUsers')

        return policy_dict['id'], policy_dict



import unittest
from unittest import mock

import pytest

from ScoutSuite.providers import get_provider
from ScoutSuite.providers.aws.authentication_strategy import AWSCredentials
from ScoutSuite.providers.base.authentication_strategy import AuthenticationException
from ScoutSuite.providers.base.authentication_strategy_factory import get_authentication_strategy
from ScoutSuite.providers.aws.resources.ec2.instances import EC2Instances

class Object(object):
    pass


# Test methods for AWS Provider
class TestAWSProviderClass(unittest.TestCase):
    @mock.patch("ScoutSuite.providers.aws.authentication_strategy.boto3")
    @mock.patch("ScoutSuite.providers.aws.authentication_strategy.get_caller_identity")
    def test_authenticate(self, mock_get_caller_identity, mock_boto3):

        aws_authentication_strategy = get_authentication_strategy("aws")

        boto3_session = Object()
        boto3_session._session = Object()
        mock_boto3.Session.return_value = boto3_session

        test_cases = [
            # no params
            {
                "profile": None,
                "aws_access_key_id": None,
                "aws_secret_access_key": None,
                "aws_session_token": None,
                "call_dict": {},
            },
            # profile
            {
                "profile": "123",
                "aws_access_key_id": None,
                "aws_secret_access_key": None,
                "aws_session_token": None,
                "call_dict": {"profile_name": "123"},
            },
            # access and secret key
            {
                "profile": None,
                "aws_access_key_id": "456",
                "aws_secret_access_key": "789",
                "aws_session_token": None,
                "call_dict": {
                    "aws_access_key_id": "456",
                    "aws_secret_access_key": "789",
                },
            },
            # access, secret key and token
            {
                "profile": None,
                "aws_access_key_id": "456",
                "aws_secret_access_key": "789",
                "aws_session_token": "101112",
                "call_dict": {
                    "aws_access_key_id": "456",
                    "aws_secret_access_key": "789",
                    "aws_session_token": "101112",
                },
            },
        ]

        for test_case in test_cases:
            result = aws_authentication_strategy.authenticate(
                test_case["profile"],
                test_case["aws_access_key_id"],
                test_case["aws_secret_access_key"],
                test_case["aws_session_token"],
            )
            mock_boto3.Session.assert_called_with(**test_case["call_dict"])
            mock_get_caller_identity.assert_called_with(boto3_session)
            assert isinstance(result, AWSCredentials)
            assert result.session == boto3_session

        # exception test
        mock_boto3.Session.side_effect = Exception("an exception")
        with pytest.raises(AuthenticationException):
            result = aws_authentication_strategy.authenticate(None, None, None, None)

    # mock two separate places from which get_aws_account_id is called
    @mock.patch("ScoutSuite.providers.aws.facade.base.get_aws_account_id")
    @mock.patch("ScoutSuite.providers.aws.facade.base.get_partition_name")
    @mock.patch("ScoutSuite.providers.aws.provider.get_aws_account_id")
    @mock.patch("ScoutSuite.providers.aws.provider.get_partition_name")
    def test_get_report_name(
            self,
            mock_get_partiton_name,
            mock_get_aws_account_id,
            mock_facade_aws_account_id,
            mock_facade_aws_partition_name,
    ):
        # no account_id, no profile
        mock_get_aws_account_id.return_value = None
        mock_get_partiton_name.return_value = None
        aws_provider = get_provider(
            provider="aws", credentials=mock.MagicMock(session="123"),
        )
        assert aws_provider.get_report_name() == "aws"

        # profile and account_id
        mock_get_aws_account_id.return_value = "12345"
        aws_provider = get_provider(
            provider="aws", profile="9999", credentials=mock.MagicMock(session="123"),
        )
        assert aws_provider.get_report_name() == "aws-9999"

        # account_id
        aws_provider = get_provider(
            provider="aws", credentials=mock.MagicMock(session="123"),
        )
        assert aws_provider.get_report_name() == "aws-12345"

    @pytest.mark.skip(reason="pytest does not reproduce actual behavior")
    def test_identify_user_data_secrets(self):

        SAMPLE_USER_DATA = """
# Various AWS Access Key exercisers
AKIASHORT # too short
AKIA0123456789ABCDEF # just right
AKIA0123456789ABCDEF0 # too long
AKIA0123456789abcdef # invalid characters
FAKIA0123456789ABCDE # wrong prefix
in middle AKIAFEDCBA9876543210 of line
line ends with AKIAFFFFFFFFFFFFFFFF

# Various AWS Secret Access Key exercisers
ThisIsTooShort
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
ThisOneIsJustALittleBitLongerThanItShouldBe
middle="0000000000/1111111111/2222222222/3333333" + "of line"
hats off to TRON: HereIsSomethingThatAppearsAtEndOfLineMCP
        """

        """
        As I write this test, the assertions below fail; somehow, the "too long"
        sequences return their initial substrings, which should not even be
        possible. This behavior appears with pytest, but not when repeated
        interactively. This behavior also does not appear with the actual scanner:

        The following is excerpted from actual (pretty-printed) output:
        [...]
        "user_data": "#!/bin/bash\ncat << \"EOF\" > /root/rsb\n# Various AWS Access Key exercisers\nAKIASHORT # too short\nAKIA0123456789ABCDEF # just right\nAKIA0123456789ABCDEF0 # too long\nAKIA0123456789abcdef # invalid characters\nFAKIA0123456789ABCDE # wrong prefix\nin middle AKIAFEDCBA9876543210 of line\nline ends with AKIAFFFFFFFFFFFFFFFF\n\n# Various AWS Secret Access Key exercisers\nThisIsTooShort\n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\nThisOneIsJustALittleBitLongerThanItShouldBe\nmiddle=\"0000000000/1111111111/2222222222/3333333\" + \"of line\"\nhats off to TRON: HereIsSomethingThatAppearsAtEndOfLineMCP\nEOF",
        "user_data_secrets": {
            "AWS Access Key IDs": [
                "AKIA0123456789ABCDEF",
                "AKIAFEDCBA9876543210",
                "AKIAFFFFFFFFFFFFFFFF"
            ],
            "AWS Secret Access Keys": [
                "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",
                "0000000000/1111111111/2222222222/3333333",
                "HereIsSomethingThatAppearsAtEndOfLineMCP"
            ]
        }
        [...]
        """

        results = EC2Instances._identify_user_data_secrets(SAMPLE_USER_DATA)
        assert results["AWS Access Key IDs"] == [
            "AKIA0123456789ABCDEF",
            "AKIAFEDCBA9876543210",
            "AKIAFFFFFFFFFFFFFFFF"
        ]
        assert results["AWS Secret Access Keys"] == [
            "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",
            "0000000000/1111111111/2222222222/3333333",
            "HereIsSomethingThatAppearsAtEndOfLineMCP"
        ]



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class Queues(AzureResources):

    def __init__(self, facade: AzureFacade, resource_group_name: str, storage_account_name: str, subscription_id: str):
        super().__init__(facade)
        self.resource_group_name = resource_group_name
        self.storage_account_name = storage_account_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        raw_queues = await self.facade.storageaccounts.get_queues(self.resource_group_name,
                                                                  self.storage_account_name,
                                                                  self.subscription_id)
        for raw_queue in raw_queues:
            id, queue = self._parse_queue(raw_queue)
            self[id] = queue

    def _parse_queue(self, raw_queue):
        pass



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class RegulatoryComplianceResults(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_regulatory_compliance_result in await \
                self.facade.securitycenter.get_regulatory_compliance_results(self.subscription_id):
            id, regulatory_compliance_result = \
                self._parse_regulatory_compliance_result(raw_regulatory_compliance_result)
            self[id] = regulatory_compliance_result

    def _parse_regulatory_compliance_result(self, raw_regulatory_compliance_result):
        regulatory_compliance_result_dict = {}
        regulatory_compliance_result_dict['id'] = get_non_provider_id(raw_regulatory_compliance_result.id)
        regulatory_compliance_result_dict['name'] = '{} {}'.format(raw_regulatory_compliance_result.standard_name,
                                                                   raw_regulatory_compliance_result.name)
        regulatory_compliance_result_dict['reference'] = raw_regulatory_compliance_result.name
        regulatory_compliance_result_dict['standard_name'] = raw_regulatory_compliance_result.standard_name
        regulatory_compliance_result_dict['type'] = raw_regulatory_compliance_result.type
        regulatory_compliance_result_dict['description'] = raw_regulatory_compliance_result.description
        regulatory_compliance_result_dict['state'] = raw_regulatory_compliance_result.state
        regulatory_compliance_result_dict['passed_assessments'] = raw_regulatory_compliance_result.passed_assessments
        regulatory_compliance_result_dict['failed_assessments'] = raw_regulatory_compliance_result.failed_assessments
        regulatory_compliance_result_dict['skipped_assessments'] = raw_regulatory_compliance_result.skipped_assessments
        regulatory_compliance_result_dict['additional_properties'] = \
            raw_regulatory_compliance_result.additional_properties
        return regulatory_compliance_result_dict['id'], regulatory_compliance_result_dict



import json
import asyncio

from ScoutSuite.core.console import print_exception, print_info, print_warning, print_debug
from ScoutSuite.providers.gcp.facade.basefacade import GCPBaseFacade
from ScoutSuite.providers.gcp.facade.cloudresourcemanager import CloudResourceManagerFacade
from ScoutSuite.providers.gcp.facade.cloudsql import CloudSQLFacade
from ScoutSuite.providers.gcp.facade.memorystoreredis import MemoryStoreRedisFacade
from ScoutSuite.providers.gcp.facade.cloudstorage import CloudStorageFacade
from ScoutSuite.providers.gcp.facade.gce import GCEFacade
from ScoutSuite.providers.gcp.facade.dns import DNSFacade
from ScoutSuite.providers.gcp.facade.iam import IAMFacade
from ScoutSuite.providers.gcp.facade.kms import KMSFacade
from ScoutSuite.providers.gcp.facade.stackdriverlogging import StackdriverLoggingFacade
from ScoutSuite.providers.gcp.facade.stackdrivermonitoring import StackdriverMonitoringFacade
from ScoutSuite.providers.gcp.facade.gke import GKEFacade
from ScoutSuite.providers.gcp.facade.functions import FunctionsFacade
from ScoutSuite.providers.gcp.facade.bigquery import BigQueryFacade
from ScoutSuite.providers.gcp.facade.utils import GCPFacadeUtils
from ScoutSuite.utils import format_service_name


class GCPFacade(GCPBaseFacade):
    def __init__(self,
                 default_project_id=None, project_id=None, folder_id=None, organization_id=None, all_projects=None):
        super().__init__('cloudresourcemanager', 'v1')

        self.default_project_id = default_project_id
        self.all_projects = all_projects
        self.project_id = project_id
        self.folder_id = folder_id
        self.organization_id = organization_id

        self.cloudresourcemanager = CloudResourceManagerFacade()
        self.cloudsql = CloudSQLFacade()
        self.cloudstorage = CloudStorageFacade()
        self.memorystoreredis = MemoryStoreRedisFacade()
        self.gce = GCEFacade()
        self.functions = FunctionsFacade()
        self.bigquery = BigQueryFacade()
        self.iam = IAMFacade()
        self.kms = KMSFacade()
        self.dns = DNSFacade()
        self.stackdriverlogging = StackdriverLoggingFacade()
        self.stackdrivermonitoring = StackdriverMonitoringFacade()

        # lock to minimize concurrent calls to get_services()
        self.projects_services_lock = False
        self.projects_services = {}

        # Instantiate facades for proprietary services
        try:
            self.gke = GKEFacade(self.gce)
        except NameError as _:
            pass

    async def get_projects(self):
        try:
            # All projects to which the user / Service Account has access to
            if self.all_projects:
                return await self._get_projects_recursively(
                    parent_type='all', parent_id=None)
            # Project passed through the CLI
            elif self.project_id:
                return await self._get_projects_recursively(
                    parent_type='project', parent_id=self.project_id)
            # Folder passed through the CLI
            elif self.folder_id:
                return await self._get_projects_recursively(
                    parent_type='folder', parent_id=self.folder_id)
            # Organization passed through the CLI
            elif self.organization_id:
                return await self._get_projects_recursively(
                    parent_type='organization', parent_id=self.organization_id)
            # Project inferred from default configuration
            elif self.default_project_id:
                return await self._get_projects_recursively(
                    parent_type='project', parent_id=self.default_project_id)
            # Raise exception if none of the above
            else:
                print_info(
                    "Could not infer the Projects to scan and no default Project ID was found.")
                return []

        except Exception as e:
            print_exception(f'Failed to retrieve projects: {e}')
            return []

    async def _get_projects_recursively(self, parent_type, parent_id):
        """
        Returns all the projects in a given organization or folder. For a project_id it only returns the project
        details.

        # FIXME can't currently be done with API client library as it consumes v1 which doesn't support folders
        resource_manager_client = resource_manager.Client(credentials=self.credentials)
        project_list = resource_manager_client.list_projects()
        for p in project_list:
            if p.parent['id'] == self.organization_id and p.status == 'ACTIVE':
                projects.append(p.project_id)
        """

        if parent_type not in ['project', 'organization', 'folder', 'all']:
            return None

        resourcemanager_client = self._get_client()
        resourcemanager_client_v2 = self._build_arbitrary_client('cloudresourcemanager', 'v2', force_new=True)

        projects = []

        try:
            projects_group = resourcemanager_client.projects()

            if parent_type == 'project':
                request = resourcemanager_client.projects().list(filter='id:"%s"' % parent_id)
            elif parent_type == 'all':
                request = resourcemanager_client.projects().list()
            # get parent children projects
            else:
                request = resourcemanager_client.projects().list(filter='parent.id:"%s"' % parent_id)

                # get parent children projects in children folders recursively
                folder_request = resourcemanager_client_v2.folders().list(parent=f'{parent_type}s/{parent_id}')
                folder_response = await GCPFacadeUtils.get_all('folders', folder_request, projects_group)
                for folder in folder_response:
                    projects.extend(await self._get_projects_recursively("folder", folder['name'].strip('folders/')))

            project_response = await GCPFacadeUtils.get_all('projects', request, projects_group)
            if project_response:
                for project in project_response:
                    if project['lifecycleState'] == "ACTIVE":
                        projects.append(project)
            else:
                print_exception('No Projects Found, '
                                'you may have specified a non-existing Organization, Folder or Project')

        except Exception as e:
            print_level = print_exception
            exception_str = str(e)
            try:
                if 'The service is currently unavailable' in exception_str or 'Internal error encountered' in exception_str:
                    print_level = print_warning
                if hasattr(e, 'content'):
                    content = e.content.decode("utf-8")
                    content_dict = json.loads(content)
                    exception_str = content_dict.get("error").get("message")
            except Exception:
                # The default output level and message have been set. Use those in the event of any error processing the exception.
                pass

            print_level(f'Unable to list accessible Projects: {exception_str}')

        finally:
            return projects

    async def get_enabled_services(self, project_id, attempt=1, has_lock=False):
        timeout = 60*attempt
        if project_id not in self.projects_services:
            # not locked, make query
            if has_lock or not self.projects_services_lock:
                self.projects_services_lock = True
                try:
                    serviceusage_client = self._build_arbitrary_client('serviceusage', 'v1', force_new=True)
                    services = serviceusage_client.services()
                    request = services.list(parent=f'projects/{project_id}', pageSize=200, filter="state:ENABLED")
                    services_response = await GCPFacadeUtils.get_all('services', request, services)
                    self.projects_services[project_id] = services_response
                    self.projects_services_lock = False
                    return self.projects_services[project_id]
                except Exception as e:
                    # hit quota, wait and retry
                    if ('API_SHARED_QUOTA_EXHAUSTED' in str(e) or 'RATE_LIMIT_EXCEEDED' in str(e)) and attempt <= 10:
                        print_warning(f"Service Usage quotas exceeded for project \"{project_id}\", retrying in {timeout}s")
                        await asyncio.sleep(timeout)
                        return await self.get_enabled_services(project_id, attempt + 1, has_lock=True)
                    # unknown error
                    else:
                        print_warning(f"Could not fetch the state of services for project \"{project_id}\": {e}")
                        self.projects_services_lock = False
                        return None
            # locked, wait and retry
            else:
                if attempt <= 10:  # need to set a limit to ensure we don't hit recursion limits
                    if attempt != 1:
                        print_debug(f"Lock already acquired for get_services() on project \"{project_id}\", retrying in {timeout}s")
                        await asyncio.sleep(timeout)
                    # set a lower threshold for the first attempt so that execution runs faster when there aren't any issues
                    else:
                        await asyncio.sleep(10)
                    return await self.get_enabled_services(project_id, attempt + 1)
                else:
                    print_warning(f"Could not fetch the state of services for project \"{project_id}\", "
                                  f"exiting before hitting maximum recursion")
                    return None
        else:
            return self.projects_services[project_id]

    async def is_api_enabled(self, project_id, service):
        """
        Given a project ID and service name, this method tries to determine if the service's API is enabled
        """

        # These are hardcoded endpoint correspondences as there's no easy way to do this.
        incorrect_endpoints = []
        # All projects have IAM policies regardless of whether the IAM API is enabled.
        if service == 'IAM':
            return True
        # These are hardcoded endpoint correspondences as there's no easy way to do this.
        elif service == 'KMS':
            endpoint = 'cloudkms'
        elif service == 'CloudStorage':
            endpoint = 'storage-component'
        elif service == 'CloudSQL':
            endpoint = 'sqladmin'
        elif service == 'ComputeEngine':
            endpoint = 'compute'
        elif service == 'Functions':
            endpoint = 'cloudfunctions'
        elif service == 'BigQuery':
            endpoint = 'bigquery'
            incorrect_endpoints.append('annotation-bigquery-public-data.cloudpartnerservices.goog')
        elif service == 'KubernetesEngine':
            endpoint = 'container'
        elif service == 'StackdriverLogging':
            endpoint = 'logging'
        elif service == 'StackdriverMonitoring':
            endpoint = 'monitoring'
        elif service == 'MemoryStore':
            endpoint = 'redis'
        elif service == 'DNS':
            endpoint = 'dns'
        else:
            print_warning(f"Could not validate the state of the {format_service_name(service.lower())} API "
                          f"for project \"{project_id}\" (unknown endpoint), including it in the execution")
            return True

        try:
            enabled_services = await self.get_enabled_services(project_id)
            if enabled_services == None:
                print_warning(f"Could not identify enabled services, including {service}")
                return True
            for s in enabled_services:
                if endpoint in s.get('name') and s.get('config').get('name') not in incorrect_endpoints:
                    print_debug(f'{format_service_name(service.lower())} API enabled for '
                                f'project \"{project_id}\", including')
                    return True
            print_info(f'{format_service_name(service.lower())} API not enabled for '
                       f'project \"{project_id}\", skipping')
            return False
        except Exception as e:
            print_warning(f"Could not validate the state of the {format_service_name(service.lower())} API "
                          f"for project \"{project_id}\": \"{e}\", including it in the execution")
            return True



from ScoutSuite.providers.base.resources.base import Resources


class NodePools(Resources):
    def __init__(self, cluster):
        super(NodePools, self).__init__(service_facade=None)
        self.cluster = cluster

    def fetch_all(self):
        raw_node_pools = self.cluster['nodePools']
        for raw_node_pool in raw_node_pools:
            node_pool_id, node_pool = self._parse_node_pool(raw_node_pool)
            self[node_pool_id] = node_pool
        # We need self.cluster to get the node pools, but we do 
        # not want to have it in the generated JSON.
        del self.cluster

    def _parse_node_pool(self, raw_node_pool):
        node_pool_dict = {}
        node_pool_dict['id'] = raw_node_pool['name']
        node_pool_dict['status'] = raw_node_pool['status']
        node_pool_dict['auto_repair_enabled'] = \
            raw_node_pool.get('management', {}).get('autoRepair', False)
        node_pool_dict['auto_upgrade_enabled'] = \
            raw_node_pool.get('management', {}).get('autoUpgrade', False)
        node_pool_dict['secure_boot_enabled'] = \
            raw_node_pool.get('config', {}).get('shieldedInstanceConfig', {}).get('enableSecureBoot', False)
        node_pool_dict['integrity_monitoring_enabled'] = \
            raw_node_pool.get('config', {}).get('shieldedInstanceConfig', {}).get('enableIntegrityMonitoring', False)
        node_pool_dict['legacy_metadata_endpoints_enabled'] = \
            raw_node_pool['config'].get('metadata', {}).get('disable-legacy-endpoints') == 'false'
        return node_pool_dict['id'], node_pool_dict




from ScoutSuite.providers.gcp.resources.projects import Projects
from ScoutSuite.providers.gcp.resources.dns.managed_zones import ManagedZones


class DNS(Projects):
    _children = [ 
        (ManagedZones, 'managed_zones')
     ]



"""This module provides implementations for Resources for Kubernetes."""

from ScoutSuite.providers.kubernetes.facade import KubernetesFacade
from ScoutSuite.providers.base.resources.base import CompositeResources, Resources


class KubernetesResources(Resources):
    """This is the base class for Kubernetes resources."""

    def __init__(self, resources):
        self.resources = resources

    async def fetch_all(self):
        for version in self.resources:
            data = self.resources[version]
            self[version] = data
            self[f'''{version}_count'''] = len(data['resources'])

class KubernetesResourcesWithFacade(Resources):
    """This is the base class for Kubernetes resources."""

    def __init__(self, facade: KubernetesFacade):
        super().__init__(facade)
        self.facade = facade
    
    def save(self, data):
        if not data: return
        for version in data:
            self[version] = data[version]

class KubernetesCompositeResources(KubernetesResourcesWithFacade, CompositeResources):
    """This class represents a collection of KubernetesResources. Classes extending KubernetesResourcesWithFacade should define a
    "_children" attribute which consists of a list of tuples describing the children. The tuples are expected to
    respect the following format: (<child_class>, <child_name>). The child_name is used by indicates the name under
    which the child will be stored in the parent object.
    """

    async def fetch_all(self):
        for child_class, child_name in self._children:
            data: KubernetesResourcesWithFacade = child_class(self.facade)
            await data.fetch_all()

            self[child_name] = {}
            for version in data:
                self[child_name][version] = 1
                self[f'{child_name}_{version}'] = data[version]
                self[f'{child_name}_{version}_count'] = len(data[version]['resources'])


from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions
from .tables import Tables


class DynamoDB(Regions):
    _children = [
        (Tables, 'tables')
    ]

    def __init__(self, facade: AWSFacade):
        super(DynamoDB, self).__init__('dynamodb', facade)




from azure.mgmt.sql import SqlManagementClient
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.core.console import print_exception
from ScoutSuite.utils import get_user_agent


class SQLDatabaseFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = SqlManagementClient(self.credentials.get_credentials(),
                                     subscription_id=subscription_id,
                                     user_agent=get_user_agent())

        return client

    async def get_database_blob_auditing_policies(self, resource_group_name, server_name, database_name,
                                                  subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: client.database_blob_auditing_policies.get(
                    resource_group_name, server_name, database_name)
            )
        except Exception as e:
            print_exception(f'Failed to retrieve database blob auditing policies: {e}')
            return []

    async def get_database_threat_detection_policies(self, resource_group_name, server_name, database_name,
                                                     subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: client.database_threat_detection_policies.get(resource_group_name, server_name, database_name,
                                                                      'default')
            )
        except Exception as e:
            print_exception(f'Failed to retrieve database threat detection policies: {e}')
            return []

    async def get_databases(self, resource_group_name, server_name, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.databases.list_by_server(resource_group_name, server_name))
            )
        except Exception as e:
            print_exception(f'Failed to retrieve databases: {e}')
            return []

    async def get_database_replication_links(self, resource_group_name, server_name, database_name,
                                             subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.replication_links.list_by_database(
                    resource_group_name, server_name, database_name))
            )
        except Exception as e:
            print_exception(f'Failed to retrieve database replication links: {e}')
            return []

    async def get_server_azure_ad_administrators(self, resource_group_name, server_name, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.server_azure_ad_administrators.list_by_server(resource_group_name, server_name))
            )
        except Exception as e:
            print_exception(f'Failed to retrieve server azure ad administrators: {e}')
            return None

    async def get_server_blob_auditing_policies(self, resource_group_name, server_name, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: client.server_blob_auditing_policies.get(resource_group_name, server_name)
            )
        except Exception as e:
            print_exception(f'Failed to retrieve server blob auditing policies: {e}')
            return []

    async def get_server_security_alert_policies(self, resource_group_name, server_name, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: client.server_security_alert_policies.get(resource_group_name, server_name, 'default')
            )
        except Exception as e:
            print_exception(f'Failed to retrieve server security alert policies: {e}')
            return []

    async def get_servers(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.servers.list())
            )
        except Exception as e:
            print_exception(f'Failed to retrieve servers: {e}')
            return []

    async def get_database_transparent_data_encryptions(self, resource_group_name, server_name, database_name,
                                                        subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: client.transparent_data_encryptions.get(
                    resource_group_name, server_name, database_name, 'current')
            )
        except Exception as e:
            print_exception(f'Failed to retrieve database transparent data encryptions: {e}')
            return []

    async def get_server_vulnerability_assessments(self, resource_group_name, server_name,
                                                   subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: client.server_vulnerability_assessments.get(resource_group_name, server_name, 'default')
            )
        except Exception as e:
            print_exception(f'Failed to retrieve server vulnerability assessments: {e}')

    async def get_server_encryption_protectors(self, resource_group_name, server_name, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: client.encryption_protectors.get(resource_group_name, server_name, 'current')
            )
        except Exception as e:
            print_exception(f'Failed to retrieve database transparent data encryptions: {e}')
            return []

    async def get_firewall_rules(self, resource_group_name, server_name, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.firewall_rules.list_by_server(resource_group_name, server_name))
            )
        except Exception as e:
            print_exception(f'Failed to retrieve firewalls rules: {e}')
            return []
